# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read a Spectrum from a dat file

These files must be in two-colum (frequency, amplitude) format
"""

from six import string_types

import numpy
from astropy.io import registry

from ...spectrum.core import Spectrum
from ... import version

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__version__ = version.version

def read_dat(filepath, fcol=0, ampcol=1, **kwargs):
    """Read a `Spectrum` from a txt file
    """
    frequency, amplitude = numpy.loadtxt(filepath, usecols=[fcol, ampcol],
                                         unpack=True)
    return Spectrum(amplitude, frequencies=frequency, **kwargs)


def identify_dat(*args, **kwargs):
    """Identify the given file as a dat file, rather than anything else

    Returns
    -------
    True
        if the filename endswith .txt or .dat
    False
        otherwise
    """
    filename = args[1][0]
    if not isinstance(filename, string_types):
        filename = filename.name
    if filename.endswith('txt') or filename.endswith('dat'):
        return True
    return False


# register this file-reader with the Spectrum class
registry.register_reader('dat', Spectrum, read_dat, force=True)
registry.register_identifier('dat', Spectrum, identify_dat)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read gravitational-wave frame (GWF) files using the LALFrame API.
"""

from __future__ import division

from astropy.io import registry

from glue.lal import CacheEntry

from .identify import register_identifier
from ....detector import Channel
from ....time import Time
from ... import (TimeSeries, StateVector, TimeSeriesDict, StateVectorDict)
from ....utils import (import_method_dependency, with_import)


@with_import('lalframe.frread')
def read_timeseries(framefile, channel, start=None, end=None, datatype=None,
                    resample=False, verbose=False, _target=TimeSeries):
    """Read a `TimeSeries` of data from a gravitational-wave frame file

    This method is a thin wrapper around `lalframe.frread.read_timeseries`
    and so can accept any input accepted by that function.

    Parameters
    ----------
    framefile : `str`, :class:`glue.lal.Cache`, :lalsuite:`LALCache`
        data source object, one of:

        - `str` : frame file path
        - :class:`glue.lal.Cache` : pure python cache object
        - :lalsuite:`LALCAche` : C-based cache object

    channel : :class:`~gwpy.detector.channel.Channel`, `str`
        data channel to read from frames
    start : `Time`, :lalsuite:`LIGOTimeGPS`, optional
        start GPS time of desired data
    end : `Time`, :lalsuite:`LIGOTimeGPS`, optional
        end GPS time of desired data
    datatype : `type`, `numpy.dtype`, `str`, optional
        identifier for desired output data type
    resample : `float`, optional
        rate of samples per second at which to resample input TimeSeries.
    verbose : `bool`, optional
        print verbose output

    Returns
    -------
    data : :class:`~gwpy.timeseries.core.TimeSeries`
        a new `TimeSeries` containing the data read from disk
    """
    lal = import_method_dependency('lal')
    # parse input arguments
    if isinstance(framefile, CacheEntry):
        framefile = framefile.path
    elif isinstance(framefile, file):
        framefile = framefile.name
    if isinstance(channel, Channel):
        channel = channel.name
    if start and isinstance(start, Time):
        start = lal.LIGOTimeGPS(start.gps)
    if end and isinstance(end, Time):
        end = lal.LIGOTimeGPS(end.gps)
    if start and end:
        duration = float(end - start)
    elif end:
        raise ValueError("If `end` is given, `start` must also be given")
    else:
        duration = None
    if start:
        try:
            start = lal.LIGOTimeGPS(start)
        except TypeError:
            start = lal.LIGOTimeGPS(float(start))
    lalts = frread.read_timeseries(framefile, channel, start=start,
                                            duration=duration,
                                            datatype=datatype, verbose=verbose)
    out = _target.from_lal(lalts)
    if resample:
        out = out.resample(resample)
    return out


@with_import('lalframe.frread')
def read_timeseriesdict(framefile, channels, **kwargs):
    """Read data for multiple channels from the given GWF-format
    ``framefile``

    Parameters
    ----------
    framefile : `str`, :class:`glue.lal.Cache`, :lalsuite:`LALCache`
        data source object, one of:

        - `str` : frame file path
        - :class:`glue.lal.Cache` : pure python cache object
        - :lalsuite:`LALCAche` : C-based cache object
    channels : `list`
        list of channel names (or `Channel` objects) to read from frame

    See Also
    --------
    :func:`~gwpy.io.gwf.lalframe.read_timeseries`
        for documentation on keyword arguments

    Returns
    -------
    dict : :class:`~gwpy.timeseries.core.TimeSeriesDict`
        dict of (channel, `TimeSeries`) data pairs
    """
    out = TimeSeriesDict()
    resample = kwargs.pop('resample', None)
    if isinstance(resample, int) or resample is None:
        resample = dict((channel, resample) for channel in channels)
    elif isinstance(resample, (tuple, list)):
        resample = dict(zip(channels, resample))
    elif not isinstance(resample, dict):
        raise ValueError("Cannot parse resample request, please review "
                         "documentation for that argument")
    for channel in channels:
        out[channel] = read_timeseries(framefile, channel,
                                       resample=resample.get(channel, None),
                                       **kwargs)
    return out


@with_import('lalframe.frread')
def read_statevector_dict(source, channels, bitss=[], **kwargs):
    """Read a `StateVectorDict` of data from a gravitational-wave
    frame file
    """
    kwargs.setdefault('_target', StateVector)
    svd = StateVectorDict(read_timeseriesdict(source, channels, **kwargs))
    for (channel, bits) in zip(channels, bitss):
        svd[channel].bits = bits
    return svd


@with_import('lalframe.frread')
def read_statevector(source, channel, bits=[], **kwargs):
    kwargs.setdefault('_target', StateVector)
    sv = read_timeseries(source, channel, **kwargs).view(StateVector)
    sv.bits = bits
    return sv


# register 'gwf' reader first
try:
    import lalframe
except ImportError:
    pass
else:
    register_identifier('gwf')
    registry.register_reader('gwf', TimeSeries, read_timeseries, force=True)
    registry.register_reader('gwf', StateVector, read_statevector, force=True)
    try:
        registry.register_reader('gwf', TimeSeriesDict, read_timeseriesdict)
    except Exception as e:
        if not str(e).startswith('Reader for format'):
            raise
    else:
        registry.register_reader('gwf', StateVectorDict, read_statevector_dict)

# register lalframe
registry.register_reader('lalframe', TimeSeries, read_timeseries)
registry.register_reader('lalframe', StateVector, read_statevector)
registry.register_reader('lalframe', TimeSeriesDict, read_timeseriesdict)
registry.register_reader('lalframe', StateVectorDict, read_statevector_dict)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Miscellaneous utilties for GWpy
"""

from __future__ import print_function

from sys import stdout

from .. import version
__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version

from .docstring import update_docstrings
from .deps import *


def gprint(*values, **kwargs):
    kwargs.setdefault('file', stdout)
    file_ = kwargs['file']
    print(*values, **kwargs)
    file_.flush()

gprint.__doc__ = print.__doc__



#!/usr/bin/env python

# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""GWpy Example: plotting a state-vector

Problem
-------

I would like to examine the state of the internal seismic isolation system supporting the Fabry-Perot mirror at the end of the Y-arm at LHO, in order to investigate a noise source.

These data are private to the LIGO Scientific Collaboration and the Virgo Collaboration, but collaboration members can use the NDS2 service to download data.
"""

from gwpy import version
__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__version__ = version.version

from gwpy.timeseries import StateVector

# define bitmask
bits = [
    'Summary state',
    'State 1 damped',
    'Stage 1 isolated',
    'Stage 2 damped',
    'Stage 2 isolated',
    'Master switch ON',
    'Stage 1 WatchDog OK',
    'Stage 2 WatchDog OK',
]

# get data
data = StateVector.fetch('L1:ISI-ETMX_ODC_CHANNEL_OUT_DQ', 'May 22 2014 14:00', 'May 22 15:00', bits=bits)
data = data.resample(16)

# make a plot
plot = data.plot(add_label='inset')
plot.set_title('LLO ETMX internal seismic isolation state')
plot.add_bitmask('0b11101110')

if __name__ == '__main__':
    try:
        outfile = __file__.replace('.py', '.png')
    except NameError:
        pass
    else:
        plot.save(outfile)
        print("Example output saved as\n%s" % outfile)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""This module provides LIGO.ORG authenticated HTML queries
"""

import os
import stat
import urllib2
import cookielib
import warnings
import tempfile
import getpass

from glue.auth.saml import HTTPNegotiateAuthHandler

from .. import version

__author__ = "Scott Koranda <scott.koranda@ligo.org>"
__version__ = version.version

COOKIE_JAR = os.path.join(tempfile.gettempdir(), getpass.getuser())
LIGO_LOGIN_URL = 'login.ligo.org'


def request(url, debug=False):
    """Request the given URL using LIGO.ORG SAML authentication.

    This requires an active Kerberos ticket for the user, to get one:

    .. code-block:: bash

       kinit albert.einstein@LIGO.ORG

    Parameters
    ----------
    url : `str`
        URL path for request
    debug : `bool`, optional
        Query in verbose debugging mode, default `False`
    """
    # set debug to 1 to see all HTTP(s) traffic
    debug = int(debug)

    # need an instance of HTTPS handler to do HTTPS
    httpshandler = urllib2.HTTPSHandler(debuglevel=debug)

    # use a cookie jar to store session cookies
    jar = cookielib.LWPCookieJar()

    # if a cookie jar exists open it and read the cookies
    # and make sure it has the right permissions
    if os.path.exists(COOKIE_JAR):
        os.chmod(COOKIE_JAR, stat.S_IRUSR | stat.S_IWUSR)
        # set ignore_discard so that session cookies are preserved
        try:
            jar.load(COOKIE_JAR, ignore_discard=True)
        except cookielib.LoadError as e:
            warnings.warn('cookielib.LoadError caught: %s' % str(e))

    # create a cookie handler from the cookie jar
    cookiehandler = urllib2.HTTPCookieProcessor(jar)
    # need a redirect handler to follow redirects
    redirecthandler = urllib2.HTTPRedirectHandler()

    # need an auth handler that can do negotiation.
    # input parameter is the Kerberos service principal.
    auth_handler = HTTPNegotiateAuthHandler(
        service_principal='HTTP@%s' % LIGO_LOGIN_URL)

    # create the opener.
    opener = urllib2.build_opener(auth_handler, cookiehandler, httpshandler,
                                  redirecthandler)

    # prepare the request object
    req = urllib2.Request(url)

    # use the opener and the request object to make the request.
    response = opener.open(req)

    # save the session cookies to a file so that they can
    # be used again without having to authenticate
    jar.save(COOKIE_JAR, ignore_discard=True)

    return response



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""GWpy tests.
"""

# XXX: HACK for astropy breaking test suite
from astropy.utils.data import _deltemps
import atexit
try:
    idx = zip(*atexit._exithandlers)[0].index(_deltemps)
except ValueError:
    pass
else:
    atexit._exithandlers.pop(idx)

# run tests
from . import (timeseries, detector)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read gravitational-wave data from a cache of files
"""

from __future__ import division

import os
from decimal import _Infinity as infinity
from math import ceil
from multiprocessing import (Process, Queue as ProcessQueue)

from astropy.io import registry

from glue.lal import Cache

from ...segments import (Segment, SegmentList)
from .. import (TimeSeries, TimeSeriesList, TimeSeriesDict,
                StateVector, StateVectorDict)


def read_cache(cache, channel, start=None, end=None, resample=None,
               nproc=1, **kwargs):
    """Read a `TimeSeries` from a cache of data files using
    multiprocessing.

    The inner-workings are agnostic of data-type, but can only handle a
    single data type at a time.

    Parameters
    ----------
    cache : :class:`glue.lal.Cache`, `str`
        cache of GWF frame files, or path to a LAL-format cache file
        on disk
    channel : :class:`~gwpy.detector.channel.Channel`, `str`
        data channel to read from frames
    start : `Time`, :lalsuite:`LIGOTimeGPS`, optional
        start GPS time of desired data
    end : `Time`, :lalsuite:`LIGOTimeGPS`, optional
        end GPS time of desired data
    resample : `float`, optional
        rate (samples per second) to resample
    format : `str`, optional
        name of data file format, e.g. ``gwf`` or ``hdf``.
    nproc : `int`, default: ``1``
        maximum number of independent frame reading processes, default
        is set to single-process file reading.

    Notes
    -----
    The number of independent processes spawned by this function can be
    calculated as ``min(maxprocesses, len(cache)//minprocesssize)``.

    Returns
    -------
    data : :class:`~gwpy.timeseries.core.TimeSeries`
        a new `TimeSeries` containing the data read from disk
    """
    cls = kwargs.pop('target', TimeSeries)
    # open cache from file if given
    if isinstance(cache, (unicode, str, file)):
        cache = open_cache(cache)

    # fudge empty cache
    if len(cache) == 0:
        return cls([], channel=channel, epoch=start)

    # use cache to get start end times
    cache.sort(key=lambda ce: ce.segment[0])
    if start is None:
        start = cache[0].segment[0]
    if end is None:
        end = cache[-1].segment[1]

    # get span
    span = Segment(start, end)
    if cls not in (StateVector, StateVectorDict) and resample:
        cache = cache.sieve(segment=span.protract(8))
    else:
        cache = cache.sieve(segment=span)
    cspan = Segment(cache[0].segment[0], cache[-1].segment[1])

    # if reading one channel, try to use lalframe, its faster
    if (isinstance(channel, str) or
           (isinstance(channel, (list, tuple)) and len(channel) == 1)):
        try:
            from lalframe import frread
        except ImportError:
            format_ = 'gwf'
        else:
            kwargs.pop('type', None)
            format_ = 'lalframe'
    # otherwise use the file extension as the format
    else:
        format_ = os.path.splitext(cache[0].path)[1][1:]

    # force one frame per process minimum
    nproc = min(nproc, len(cache))

    # single-process
    if nproc <= 1:
        return cls.read(cache, channel, format=format_, start=start, end=end,
                        resample=resample, **kwargs)

    # define how to read each frame
    def _read(q, pstart, pend):
        # don't go beyond the requested limits
        pstart = float(max(start, pstart))
        pend = float(min(end, pend))
        # if resampling TimeSeries, pad by 8 seconds inside cache limits
        if cls not in (StateVector, StateVectorDict) and resample:
            cstart = float(max(cspan[0], pstart - 8))
            subcache = cache.sieve(segment=Segment(cstart, pend))
            out = cls.read(subcache, channel, format=format_, start=cstart,
                           end=pend, resample=None, **kwargs)
            out = out.resample(resample)
            q.put(out.crop(pstart, pend))
        else:
            subcache = cache.sieve(segment=Segment(pstart, pend))
            q.put(cls.read(subcache, channel, format=format_, start=pstart,
                           end=pend, resample=resample, **kwargs))

    # separate cache into parts
    fperproc = int(ceil(len(cache) / nproc))
    subcaches = [Cache(cache[i:i+fperproc]) for
                 i in range(0, len(cache), fperproc)]
    subsegments = SegmentList([Segment(c[0].segment[0], c[-1].segment[1])
                               for c in subcaches])

    # start all processes
    queue = ProcessQueue(nproc)
    proclist = []
    for subseg in subsegments:
        process = Process(target=_read, args=(queue, subseg[0], subseg[1]))
        process.daemon = True
        proclist.append(process)
        process.start()

    # get data and block
    data = [queue.get() for p in proclist]
    for process in proclist:
        process.join()

    # format and return
    if issubclass(cls, dict):
        try:
            data.sort(key=lambda tsd: tsd.values()[0].epoch.gps)
        except IndexError:
            pass
        out = cls()
        while len(data):
            tsd = data.pop(0)
            out.append(tsd)
            del tsd
        return out
    else:
        out = TimeSeriesList(*data)
        out.sort(key=lambda ts: ts.epoch.gps)
        ts = out.join()
        return ts


def read_state_cache(*args, **kwargs):
    kwargs.setdefault('target', StateVector)
    return read_cache(*args, **kwargs)


def read_dict_cache(*args, **kwargs):
    kwargs.setdefault('target', TimeSeriesDict)
    return read_cache(*args, **kwargs)


def read_state_dict_cache(*args, **kwargs):
    kwargs.setdefault('target', StateVectorDict)
    return read_cache(*args, **kwargs)


def open_cache(lcf):
    """Read a LAL-format cache file into memory as a
    :class:`glue.lal.Cache`.
    """
    if isinstance(lcf, file):
        return Cache.fromfile(lcf)
    else:
        with open(lcf, 'r') as f:
            return Cache.fromfile(f)


def identify_cache_file(*args, **kwargs):
    """Determine an input object as either a LAL-format cache file.
    """
    cachefile = args[1]
    if isinstance(cachefile, file):
        cachefile = cachefile.name
    # identify string
    if (isinstance(cachefile, (unicode, str)) and
            (cachefile.endswith('.lcf') or cachefile.endswith('.cache'))):
        return True
        # identify cache object
    else:
        return False


def identify_cache(*args, **kwargs):
    """Determine an input object as a :class:`glue.lal.Cache` or a
    :lalsuite:`LALCache`.
    """
    cacheobj = args[3]
    if isinstance(cacheobj, Cache):
            return True
    try:
        from lal import Cache as LALCache
    except ImportError:
        pass
    else:
        if isinstance(cacheobj, LALCache):
            return True
    return False


registry.register_reader('lcf', TimeSeries, read_cache)
registry.register_reader('cache', TimeSeries, read_cache)
registry.register_identifier('lcf', TimeSeries, identify_cache_file)
registry.register_identifier('cache', TimeSeries, identify_cache)

# duplicate for state-vector
registry.register_reader('lcf', StateVector, read_state_cache)
registry.register_reader('cache', StateVector, read_state_cache)
registry.register_identifier('lcf', StateVector, identify_cache_file)
registry.register_identifier('cache', StateVector, identify_cache)

# TimeSeriesDict
registry.register_reader('lcf', TimeSeriesDict, read_dict_cache)
registry.register_reader('cache', TimeSeriesDict, read_dict_cache)
registry.register_reader('lcfmp', TimeSeriesDict, read_dict_cache)
registry.register_identifier('lcf', TimeSeriesDict, identify_cache_file)
registry.register_identifier('cache', TimeSeriesDict, identify_cache)

# StateVectorDict
registry.register_reader('lcf', StateVectorDict, read_state_dict_cache)
registry.register_reader('cache', StateVectorDict, read_state_dict_cache)
registry.register_reader('lcfmp', StateVectorDict, read_state_dict_cache)
registry.register_identifier('lcf', StateVectorDict, identify_cache_file)
registry.register_identifier('cache', StateVectorDict, identify_cache)



#!/usr/bin/env python

# Copyright (C) Duncan Macleod (2014)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""GWpy Example: plotting the rate versus time of a table of event triggers

Problem
-------

I would like to study the rate at which event triggers are generated by the
`ExcessPower` gravitational-wave burst detection algorithm, over a small
stretch of data, binned by various thresholds on signal-to-noise ratio (SNR).

The data from which these events were generated are a simulation of Gaussian noise
with the Advanced LIGO design spectrum, and so don't actually contain any real
gravitational waves, but will help tune the algorithm to improve detection of
future, real signals.
"""

from gwpy import version
__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__version__ = version.version

from gwpy.table.lsctables import SnglBurstTable

# read triggers
events = SnglBurstTable.read('../gwpy/tests/data/'
                             'H1-LDAS_STRAIN-968654552-10.xml.gz')

# calculate event rate by SNR
rates = events.binned_event_rates(1, 'snr', [2, 3, 5, 8], operator='>=',
                                  start=968654552, end=968654562)

# make a plot
plot2 = rates.plot(label='name')
plot2.set_xlim(968654552, 968654562)
plot2.set_ylabel('Event rate [Hz]')
plot2.set_title('LIGO Hanford Observatory event rate for GW100916')
plot2.add_legend()

if __name__ == '__main__':
    try:
        outfile = __file__.replace('.py', '.png')
    except NameError:
        pass
    else:
        plot2.save(outfile)
        print("Example output saved as\n%s" % outfile)



# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Parsing for mathematical operations given on the command-line
"""

from ..version import version

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version

import operator
from StringIO import StringIO
import token
from tokenize import generate_tokens

OPERATORS = {'<': operator.lt,
             '<=': operator.le,
             '=': operator.eq,
             '>=': operator.ge,
             '>': operator.gt,
             '==': operator.is_,
             '!=': operator.is_not}
INV_OPERATORS = {'<': operator.ge,
                 '<=': operator.gt,
                 '=': operator.ne,
                 '>=': operator.lt,
                 '>': operator.le,
                 '==': operator.is_not,
                 '!=': operator.is_}


def parse_math_operation(opstring):
    """Parse a mathematical operation string into a list of variable,
    one or two operators, and matching thresholds thresholds
    """
    splitcon = [tok[0:2] for tok in
                generate_tokens(StringIO(opstring).readline) if tok[1]]
    # parse variable name
    variables = [a[1] for a in splitcon if a[0] in [token.STRING, token.NAME]]
    if len(variables) != 1:
        raise ValueError("Found multiple variables in operation string, "
                         "cannot parse")
    # parse limits
    limits = [a[1] for a in splitcon if token.tok_name[a[0]] == 'NUMBER']
    # parse operators
    operators = [a[1] for a in splitcon if token.tok_name[a[0]] == 'OP']

    # test correct syntax
    if len(limits) != len(operators):
        raise ValueError("Extracted %s limits and %s conditions, please "
                         "reenter" % (len(limits), len(operators)))

    # re-direct operator to put the variable on the left
    for i in range(len(limits)):
        # test direction
        if opstring.find(limits[i]) < opstring.find(operators[i]):
            operators[i] = INV_OPERATORS[operators[i]]
        else:
            operators[i] = OPERATORS[operators[i]]

    # try to format the limits
    for i,l in enumerate(limits):
        try:
            limits[i] = float(l)
        except ValueError:
            pass

    return variables[0], zip(operators, limits)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""An extension of the Plot class for handling Spectrograms
"""

import re
import numpy
import math

from ..time import Time
from ..segments import SegmentList
from .timeseries import TimeSeriesPlot

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__all__ = ['SpectrogramPlot']


class SpectrogramPlot(TimeSeriesPlot):
    """`Figure` for displaying a :class:`~gwpy.spectrogram.core.Spectrogram`.
    """
    def __init__(self, *spectrograms, **kwargs):
        self._logy = False

        # extract plotting keyword arguments
        plotargs = dict()
        plotargs['vmin'] = kwargs.pop('vmin', None)
        plotargs['vmax'] = kwargs.pop('vmax', None)
        plotargs['norm'] = kwargs.pop('norm', None)
        plotargs['cmap'] = kwargs.pop('cmap', None)
        sep = kwargs.pop('sep', True)

        # initialise figure
        super(SpectrogramPlot, self).__init__(**kwargs)

        # plot data
        for i,spectrogram in enumerate(spectrograms):
            self.add_spectrogram(spectrogram, newax=sep, **plotargs)
            self.axes[-1].set_yscale('log')
            self.axes[-1].fmt_ydata = lambda f: ('%s %s'
                                                 % (f, spectrogram.yunit))
            self.axes[-1].set_ylabel('Frequency [%s]' % spectrogram.yunit)

        # set matching epoch for each set of axes
        if len(spectrograms) and not sep:
            span = SegmentList([spec.span for spec in spectrograms]).extent()
            for ax in self.axes:
                ax.set_xlim(*span)
            for ax in self.axes[:-1]:
                ax.set_xlabel("")



#!/usr/bin/env python

# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""GWpy Example: plotting a spectrogram

Problem
-------

I would like to study the gravitational wave strain spectrogram around the time of an interesting simulated signal during the last science run (S6). I have access to the frame files on the LIGO Data Grid machine `ldas-pcdev2.ligo-wa.caltech.edu` and so can read them directly.
"""

from gwpy import version
__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__version__ = version.version

# import the TimeSeries class
from gwpy.timeseries import TimeSeries

# find the data using NDS
gwdata = TimeSeries.fetch('H1:LDAS-STRAIN', 'September 16 2010 06:40', 'September 16 2010 06:50')
gwdata.unit = 'strain'

# calculate spectrogram
specgram = gwdata.spectrogram(5, fftlength=2, overlap=1) ** (1/2.)
medratio = specgram.ratio('median')

# plot
plot = medratio.plot(norm='log', vmin=0.1, vmax=10)
plot.set_ylim(40, 4096)
plot.add_colorbar(label='Amplitude relative to median')

if __name__ == '__main__':
    try:
        outfile = __file__.replace('.py', '.png')
    except NameError:
        pass
    else:
        plot.save(outfile)
        print("Example output saved as\n%s" % outfile)



#!/usr/bin/env python

# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>

"""GWpy Example: plotting a spectrum

Problem
-------

I'm interested in the level of ground motion surrounding a particular time
during commissioning of the Advanced LIGO Livingston Observatory. I don't
have access to the frame files on disk, so I'll need to use NDS.

"""

from gwpy.time import tconvert
from gwpy.timeseries import TimeSeries

from gwpy import version
__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__version__ = version.version

# read the data over the network
lho = TimeSeries.fetch('H1:LDAS-STRAIN', 'August 1 2010', 'August 1 2010 00:02')
lho.unit = 'strain'
#llo = TimeSeries.fetch('L1:LDAS-STRAIN', start.gps, end.gps, verbose=True)
#llo.unit = 'strain'

# calculate spectrum with 0.5Hz resolution
lhoasd = lho.asd(2, 1)
#lloasd = llo.asd(2, 1)

# plot
plot = lhoasd.plot(color='b', label='LHO')
#plot.add_spectrum(lloasd, color='g', label='LLO')
plot.xlim = [40, 4096]
plot.ylim = [1e-23, 7.5e-21]

# hide some axes
plot.axes[0].spines['right'].set_visible(False)
plot.axes[0].spines['top'].set_visible(False)
plot.axes[0].get_xaxis().tick_bottom()
plot.axes[0].get_yaxis().tick_right()

if __name__ == '__main__':
    try:
        outfile = __file__.replace('.py', '.png')
    except NameError:
        pass
    else:
        plot.save(outfile)
        print("Example output saved as\n%s" % outfile)



#!/usr/bin/env python

# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>"""GWpy Example: plotting a time-series

"""
GWpy Example: plotting a transfer function
------------------------------------------

I would like to study how a signal transfers from one part of the
interferometer to another.

Specifically, it is interesting to measure the amplitude transfer of
ground motion through the HEPI system.
"""

from gwpy.time import tconvert
from gwpy.timeseries import TimeSeriesDict
from gwpy.plotter import BodePlot

from gwpy import version
__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__version__ = version.version

# set the times
start = tconvert('May 27 2014 04:00')
end = start + 1800

gndchannel = 'L1:ISI-GND_STS_ITMY_Z_DQ'
hpichannel = 'L1:HPI-ITMY_BLND_L4C_Z_IN1_DQ'

# get data
data = TimeSeriesDict.fetch([gndchannel, hpichannel], start, end, verbose=True)
gnd = data[gndchannel]
gnd.name = 'Before HEPI (ground)'
hpi = data[hpichannel]
hpi.name = 'After HEPI'

gnd.unit = 'nm/s'
hpi.unit = 'nm/s'

# get FFTs
gndfft = gnd.average_fft(100, 50, window='hamming')
hpifft = hpi.average_fft(100, 50, window='hamming')

# get transfer function (up to lower Nyquist)
size = min(gndfft.size, hpifft.size)
tf = hpifft[:size] / gndfft[:size]

plot = BodePlot(tf)
magax = plot.axes[0]
magax.set_title(r'L1 ITMY ground $\rightarrow$ HPI transfer function')
magax.set_ylim(-55, 50)


if __name__ == '__main__':
    try:
        outfile = __file__.replace('.py', '.png')
    except NameError:
        pass
    else:
        plot.save(outfile)
        print("Example output saved as\n%s" % outfile)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""This module provides plotting utilities for visualising GW data

The standard data types (`TimeSeries`, `Table`, `DataQualityFlag`, ...) can
all be easily visualised using the relevant plotting objects, with
many configurable parameters both interactive, and in saving to disk.
"""

from matplotlib import (rcParams, pyplot)

from .. import version

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__version__ = version.version

from .tex import USE_TEX
from .gps import *
from .log import *

from .core import *
from .timeseries import *
from .spectrogram import *
from .spectrum import *
from .segments import *
from .filter import *
from .table import *
from .histogram import *

GWPY_PLOT_PARAMS = {
    "axes.color_cycle": ['b', 'g', 'r', 'c', 'm', 'y', 'gray', 'gold',
                         'k', 'orange'],
    "axes.grid": True,
    "axes.axisbelow": False,
    "axes.labelsize": 22,
    'axes.titlesize': 22,
    'figure.subplot.bottom': 0.13,
    'figure.subplot.left': 0.15,
    'figure.subplot.right': 0.88,
    'figure.subplot.top': 0.88,
    "image.aspect": 'auto',
    "image.interpolation": 'nearest',
    "image.origin": 'lower',
    "xtick.labelsize": 20,
    "ytick.labelsize": 20}
if rcParams['text.usetex'] or USE_TEX:
    GWPY_PLOT_PARAMS.update({"text.usetex": True, "font.family": "serif",
                             "font.serif": ["Computer Modern"]})
rcParams.update(GWPY_PLOT_PARAMS)


def figure(*args, **kwargs):
    kwargs.setdefault('FigureClass', Plot)
    return pyplot.figure(*args, **kwargs)
figure.__doc__ = pyplot.figure.__doc__



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""This module provides an extension to the :class:`numpy.ndarray`
data structure providing metadata

The `Array` structure provides the core array-with-metadata environment
with the standard array methods wrapped to return instances of itself.
"""

from copy import deepcopy

import numpy
numpy.set_printoptions(threshold=200, linewidth=65)

from astropy.units import (Unit, UnitBase, Quantity)
from ..io import (reader, writer)

from .. import version
from ..detector import Channel
from ..time import (Time, to_gps)
from ..utils import with_import

__version__ = version.version
__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__credits__ = "Nickolas Fotopoulos <nvf@gravity.phys.uwm.edu>"


# -----------------------------------------------------------------------------
# Core Array

class Array(numpy.ndarray):
    """An extension of the :class:`~numpy.ndarray`, with added
    metadata

    This `Array` holds the input data and a standard set of metadata
    properties associated with GW data.

    Parameters
    ----------
    data : array-like, optional, default: `None`
        input data array
    name : `str`, optional, default: `None`
        descriptive title for this `Array`
    unit : `~astropy.units.core.Unit`
        physical unit of these data
    epoch : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
        starting GPS time of this `Array`, accepts any input for
        :meth:`~gwpy.time.to_gps`.
    channel : `~gwpy.detector.Channel`, `str`
        source data stream for these data
    dtype : :class:`~numpy.dtype`, optional, default: `None`
        input data type
    copy : `bool`, optional, default: `False`
        choose to copy the input data to new memory
    subok : `bool`, optional, default: `True`
        allow passing of sub-classes by the array generator
    **metadata
        other metadata properties

    Returns
    -------
    array : `Array`
        a new array, with a view of the data, and all associated metadata
    """
    __array_priority_ = 10.1
    _metadata_type = dict
    _metadata_slots = ['name', 'unit', 'epoch', 'channel']

    def __new__(cls, data=None, dtype=None, copy=False, subok=True, **metadata):
        """Define a new `Array`, potentially from an existing one
        """
        # get dtype out of the front
        if isinstance(data, numpy.ndarray) and dtype is None:
            dtype = data.dtype
        elif isinstance(data, numpy.ndarray):
            dtype = numpy.dtype(dtype)

        # copy from an existing Array
        if isinstance(data, cls):
            if not copy and dtype == data.dtype and not metadata:
                return data
            elif metadata:
                new = numpy.array(data, dtype=dtype, copy=copy, subok=True)
                new.metadata = cls._metadata_type(metadata)
                return new
            else:
                new = data.astype(dtype)
                new.metadata = data.metadata
                return new
        # otherwise define a new Array from the array-like data
        else:
            _baseclass = type(data)
            if copy:
                new = super(Array, cls).__new__(cls, data.shape, dtype=dtype)
                try:
                    new[:] = numpy.array(data, dtype=dtype, copy=True)
                except ValueError:
                    pass
            else:
                new = numpy.array(data, dtype=dtype, copy=copy,
                                  subok=True).view(cls)
            new.metadata = cls._metadata_type()
            for key, val in metadata.iteritems():
                if val is not None:
                    setattr(new, key, val)
            new._baseclass = _baseclass
            return new

    # -------------------------------------------
    # array manipulations

    def __array_finalize__(self, obj):
        """Finalize a Array with metadata
        """
        self.metadata = getattr(obj, 'metadata', {}).copy()
        self._baseclass = getattr(obj, '_baseclass', type(obj))

    def __array_wrap__(self, obj, context=None):
        """Wrap an array as a Array with metadata
        """
        result = obj.view(self.__class__)
        result.metadata = self.metadata.copy()
        return result

    def __repr__(self):
        """Return a representation of this object

        This just represents each of the metadata objects appriopriately
        after the core data array
        """
        indent = ' '*len('<%s(' % self.__class__.__name__)
        array = repr(self.data)[6:-1].replace('\n'+' '*6, '\n'+indent)
        if 'dtype' in array:
            array += ','
        metadatarepr = []
        for key in self._metadata_slots:
            try:
                val = getattr(self, key)
            except (AttributeError, KeyError):
                val = None
            mindent = ' ' * (len(key) + 1)
            rval = repr(val).replace('\n', '\n%s' % (indent+mindent))
            metadatarepr.append('%s=%s' % (key, rval))
        metadata = (',\n%s' % indent).join(metadatarepr)
        return "<%s(%s\n%s%s)>" % (self.__class__.__name__, array,
                                   indent, metadata)

    def __str__(self):
        """Return a printable string format representation of this object

        This just prints each of the metadata objects appropriately
        after the core data array
        """
        indent = ' '*len('%s(' % self.__class__.__name__)
        array = str(self.data).replace('\n', '\n' + indent) + ','
        if 'dtype' in array:
            array += ','
        metadatarepr = []
        for key in self._metadata_slots:
            try:
                val = getattr(self, key)
            except (AttributeError, KeyError):
                val = None
            if key == 'epoch' and val is not None:
                val = self.epoch.iso
            elif not val:
                val = None
            mindent = ' ' * (len(key) + 1)
            rval = str(val).replace('\n', '\n%s' % (indent+mindent))
            metadatarepr.append('%s: %s' % (key, rval))
        metadata = (',\n%s' % indent).join(metadatarepr)
        return "%s(%s\n%s%s)" % (self.__class__.__name__, array,
                                 indent, metadata)

    # -------------------------------------------
    # array methods

    def __pow__(self, y, z=None):
        new = self.copy()
        numpy.power(self, y, out=new)
        new.unit = self.unit.__pow__(y)
        return new
    __pow__.__doc__ = numpy.ndarray.__pow__.__doc__

    def __ipow__(self, y):
        super(Array, self).__ipow__(y)
        self.unit **= y
        return self
    __ipow__.__doc__ = numpy.ndarray.__ipow__.__doc__

    def median(self, axis=None, out=None, overwrite_input=False):
        return numpy.median(self, axis=axis, out=out,
                            overwrite_input=overwrite_input)
    median.__doc__ = numpy.median.__doc__

    @property
    def T(self):
        return self.transpose()

    @property
    def H(self):
        return self.T.conj()

    @property
    def data(self):
        return self.view(numpy.ndarray)
    A = data

    def copy(self, order='C'):
        new = super(Array, self).copy(order=order)
        new.metadata = deepcopy(self.metadata)
        return new
    copy.__doc__ = numpy.ndarray.copy.__doc__

    # -------------------------------------------
    # Pickle helpers

    def __getstate__(self):
        """Return the internal state of the object.

        Returns
        -------
        state : `tuple`
            A 5-tuple of (shape, dtype, typecode, rawdata, metadata)
            for pickling
        """
        state = (self.shape,
                 self.dtype,
                 self.flags.fnc,
                 self.data.tostring(),
                 self.metadata
                 )
        return state

    def __setstate__(self, state):
        """Restore the internal state of the `Array`.

        This is used for unpickling purposes.

        Parameters
        ----------
        state : `tuple`
            typically the output of the :meth:`Array.__getstate__`
            method, a 5-tuple containing:

            - class name
            - a tuple giving the shape of the data
            - a typecode for the data
            - a binary string for the data
            - the metadata dict
        """
        (shp, typ, isf, raw, meta) = state
        super(Array, self).__setstate__((shp, typ, isf, raw))
        self.metadata = self._metadata_type(meta)

    def __reduce__(self):
        """Initialise the pickle operation for this `Array`

        Returns
        -------
        pickler : `tuple`
            A 3-tuple of (reconstruct function, reconstruct args, state)
        """
        return (_array_reconstruct, (self.__class__, self.dtype),
                self.__getstate__())

    # -------------------------------------------
    # Array properties

    @property
    def name(self):
        """Name for this `Array`

        :type: `str`
        """
        try:
            return self.metadata['name']
        except KeyError:
            return None

    @name.setter
    def name(self, val):
        self.metadata['name'] = str(val)

    @property
    def unit(self):
        """Unit for this `Array`

        :type: :class:`~astropy.units.core.Unit`
        """
        try:
            return self.metadata['unit']
        except KeyError:
            self.unit = ''
            return self.unit

    @unit.setter
    def unit(self, val):
        if val is None or isinstance(val, Unit):
            self.metadata['unit'] = val
        else:
            self.metadata['unit'] = Unit(val)

    @property
    def epoch(self):
        """Starting GPS time epoch for this `Array`.

        This attribute is recorded as a `~gwpy.time.Time` object in the
        GPS format, allowing native conversion into other formats.

        See `~astropy.time` for details on the `Time` object.
        """
        try:
            return Time(float(self.metadata['epoch']), format='gps', scale='utc')
        except KeyError:
            return None

    @epoch.setter
    def epoch(self, epoch):
        self.metadata['epoch'] = to_gps(epoch)

    @property
    def channel(self):
        """Data channel associated with this `Array`.
        """
        try:
            return self.metadata['channel']
        except KeyError:
            return None

    @channel.setter
    def channel(self, ch):
        if isinstance(ch, Channel):
            self.metadata['channel'] = ch
        else:
            self.metadata['channel'] = Channel(ch)

    # -------------------------------------------
    # extras

    @classmethod
    def _getAttributeNames(cls):
        return cls._metadata_slots

    # use input/output registry to allow multi-format reading
    read = classmethod(reader())
    write = writer()

    @with_import('h5py')
    def to_hdf5(self, output, name=None, group=None, compression='gzip',
                **kwargs):
        """Convert this `Array` to a :class:`h5py.Dataset`.

        This allows writing to an HDF5-format file.

        Parameters
        ----------
        output : `str`, :class:`h5py.Group`
            path to new output file, or open h5py `Group` to write to.
        name : `str`, optional
            custom name for this `Array` in the HDF hierarchy, defaults
            to the `name` attribute of the `Array`.
        group : `str`, optional
            group to create for this time-series.
        compression : `str`, optional
            name of compression filter to use
        **kwargs
            other keyword arguments passed to
            :meth:`h5py.Group.create_dataset`.

        Returns
        -------
        dset : :class:`h5py.Dataset`
            HDF dataset containing these data.
        """
        # create output object
        if isinstance(output, h5py.Group):
            h5file = output
        else:
            h5file = h5py.File(output, 'w')

        try:
            # if group
            if group:
                try:
                    h5group = h5file[group]
                except KeyError:
                    h5group = h5file.create_group(group)
            else:
                h5group = h5file

            # create dataset
            name = name or self.name
            if name is None:
                raise ValueError("Cannot store Array without a name. "
                                 "Either assign the name attribute of the "
                                 "array itself, or given name= as a keyword "
                                 "argument to write().")
            dset = h5group.create_dataset(name or self.name, data=self.data,
                                          compression=compression,
                                          **kwargs)

            # store metadata
            for attr, mdval in self.metadata.iteritems():
                if isinstance(mdval, Quantity):
                    dset.attrs[attr] = mdval.value
                elif isinstance(mdval, Channel):
                    dset.attrs[attr] = mdval.ndsname
                elif isinstance(mdval, UnitBase):
                    dset.attrs[attr] = str(mdval)
                else:
                    dset.attrs[attr] = mdval

        finally:
            if not isinstance(output, h5py.Group):
                h5file.close()

        return dset

    @classmethod
    @with_import('h5py')
    def from_hdf5(cls, f, name=None):
        """Read a `Array` from the given HDF file.

        Parameters
        ----------
        f : `str`, :class:`h5py.HLObject`
            path to HDF file on disk, or open `h5py.HLObject`.
        type_ : `type`
            target class to read
        name : `str`
            path in HDF hierarchy of dataset.
        """
        from ..io.hdf5 import open_hdf5

        h5file = open_hdf5(f)

        try:
            # find dataset
            if isinstance(h5file, h5py.Dataset):
                dataset = h5file
            else:
                try:
                    dataset = h5file[name]
                except KeyError:
                    if name.startswith('/'):
                        raise
                    name2 = '/%s/%s' % (cls.__name__.lower(), name)
                    if name2 in h5file:
                        dataset = h5file[name2]
                    else:
                        raise

            # read array, close file, and return
            out = cls(dataset[()], **dict(dataset.attrs))
        finally:
            if not isinstance(f, (h5py.Dataset, h5py.Group)):
                h5file.close()

        return out


def _array_reconstruct(class_, dtype):
    """Reconstruct an `Array` after unpickling

    Parameters
    ----------
    Class : `type`, `Array` or sub-class
        class object to create
    dtype : `type`, `numpy.dtype`
        dtype to set
    """
    return class_.__new__(class_, [], dtype=dtype)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Wrapper to the nds2-client package, providing network access
to LIGO data.
"""

import sys

import nds2

from .. import version
from .kerberos import *

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__version__ = version.version

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

DEFAULT_HOSTS = OrderedDict([
    (None, ('ldas-pcdev4.ligo.caltech.edu', 31200)),
    (None, ('nds.ligo.caltech.edu', 31200)),
    ('H1', ('nds.ligo-wa.caltech.edu', 31200)),
    ('H0', ('nds.ligo-wa.caltech.edu', 31200)),
    ('L1', ('nds.ligo-la.caltech.edu', 31200)),
    ('L0', ('nds.ligo-la.caltech.edu', 31200)),
    ('C1', ('nds40.ligo.caltech.edu', 31200)),
    ('C0', ('nds40.ligo.caltech.edu', 31200))])

# set type dicts
NDS2_CHANNEL_TYPESTR = {}
for ctype in (nds2.channel.CHANNEL_TYPE_RAW,
              nds2.channel.CHANNEL_TYPE_ONLINE,
              nds2.channel.CHANNEL_TYPE_RDS,
              nds2.channel.CHANNEL_TYPE_STREND,
              nds2.channel.CHANNEL_TYPE_MTREND,
              nds2.channel.CHANNEL_TYPE_STATIC,
              nds2.channel.CHANNEL_TYPE_TEST_POINT):
    NDS2_CHANNEL_TYPESTR[ctype] = nds2.channel_channel_type_to_string(ctype)
NDS2_CHANNEL_TYPE = dict((val, key) for (key, val) in
                         NDS2_CHANNEL_TYPESTR.iteritems())


class NDSOutputContext(object):
    def __init__(self, stdout=sys.stdout, stderr=sys.stderr):
        self._stdout = stdout or sys.stdout
        self._stderr = stderr or sys.stderr

    def __enter__(self):
        self.old_stdout, self.old_stderr = sys.stdout, sys.stderr
        self.old_stdout.flush()
        self.old_stderr.flush()
        sys.stdout, sys.stderr = self._stdout, self._stderr

    def __exit__(self, *args):
        self._stdout.flush()
        self._stderr.flush()
        sys.stdout = self.old_stdout
        sys.stderr = self.old_stderr


class NDSWarning(UserWarning):
    pass


def host_resolution_order(ifo):
    hosts = []
    if ifo in DEFAULT_HOSTS:
        hosts.append(DEFAULT_HOSTS[ifo])
    for difo, hp in DEFAULT_HOSTS.iteritems():
        if difo != ifo and hp not in hosts:
            hosts.append(hp)
    return hosts



"""
:Description: Sphinx extension to remove leading under-scores from directories names in the html build output directory.
"""
import os
import shutil


def setup(app):
    """
    Add a html-page-context  and a build-finished event handlers
    """
    app.connect('html-page-context', change_pathto)
    app.connect('build-finished', move_private_folders)
                
def change_pathto(app, pagename, templatename, context, doctree):
    """
    Replace pathto helper to change paths to folders with a leading underscore.
    """
    pathto = context.get('pathto')
    def gh_pathto(otheruri, *args, **kw):
        if otheruri.startswith('_'):
            otheruri = otheruri[1:]
        return pathto(otheruri, *args, **kw)
    context['pathto'] = gh_pathto
    
def move_private_folders(app, e):
    """
    remove leading underscore from folders in in the output folder.
    
    :todo: should only affect html built
    """
    def join(dir):
        return os.path.join(app.builder.outdir, dir)
    
    for item in os.listdir(app.builder.outdir):
        if item.startswith('_') and os.path.isdir(join(item)):
            shutil.move(join(item), join(item[1:]))


# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Input/output routines for gravitational-wave frame (GWF) format files.

The frame format is defined in LIGO-T970130 available from dcc.ligo.org.

Currently supported are two separate libraries:

- `lalframe` : using the LIGO Algorithm Library Frame API (based off the
  FrameL library)
- `framecpp` : using the alternative ``frameCPP`` library

Due to the lower-level nature of the frameCPP python package, it is
preferred, in the instance that both lalframe and frameCPP are available
on a system.
"""

from ....version import version

from . import lalfr
from . import framecpp

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit test for timeseries module
"""

import os
import os.path
import unittest

from numpy import random

from astropy import units

from gwpy.time import Time

from gwpy import version
from gwpy.timeseries import TimeSeries

SEED = 1
GPS_EPOCH = Time(0, format='gps', scale='utc')
ONE_HZ = units.Quantity(1, 'Hz')
ONE_SECOND = units.Quantity(1, 'second')

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version


class TimeSeriesTests(unittest.TestCase):
    """`TestCase` for the timeseries module
    """
    framefile = os.path.join(os.path.split(__file__)[0], 'data',
                             'HLV-GW100916-968654552-1.gwf')

    def setUp(self):
        random.seed(SEED)
        self.data = random.random(100)

    def test_creation(self):
        TimeSeries(self.data)

    def test_creation_with_metadata(self):
        self.ts = TimeSeries(self.data, sample_rate=1, name='TEST CASE',
                             epoch=0, channel='TEST CASE')
        repr(self.ts)
        self.assertTrue(self.ts.epoch == GPS_EPOCH)
        self.assertTrue(self.ts.sample_rate == ONE_HZ)
        self.assertTrue(self.ts.dt == ONE_SECOND)

    def frame_read(self, format=None):
        ts = TimeSeries.read(self.framefile, 'L1:LDAS-STRAIN', format=format)
        self.assertTrue(ts.epoch == Time(968654552, format='gps',
                                              scale='utc'))
        self.assertTrue(ts.sample_rate == units.Quantity(16384, 'Hz'))
        self.assertTrue(ts.unit == units.Unit('strain'))

    def test_frame_read_lalframe(self):
        try:
            self.frame_read(format='lalframe')
        except ImportError as e:
            raise unittest.SkipTest(str(e))

    def test_frame_read_framecpp(self):
        try:
            self.frame_read(format='framecpp')
        except ImportError as e:
            raise unittest.SkipTest(str(e))


if __name__ == '__main__':
    unittest.main()



# coding=utf-8
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""The `Series` is a one-dimensional array with metadata
"""

import math
import inspect

import numpy
from scipy import signal
from astropy.units import (Unit, Quantity)

from .. import version
__version__ = version.version
__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

from .array import Array
from .series import Series
from ..segments import Segment


class Array2D(Array):
    """A two-dimensional array with metadata
    """
    _metadata_slots = (Array._metadata_slots +
                       ['x0', 'dx', 'y0', 'dy', 'logx', 'logy'])
    xunit = Unit('')
    yunit = Unit('')

    def __new__(cls, data, dtype=None, copy=False, subok=True, **metadata):
        """Define a new `Array2D`
        """
        if len(data) and not numpy.asarray(data).ndim == 2:
            raise ValueError("Data must be two-dimensional")
        return super(Array2D, cls).__new__(cls, data, dtype=dtype, copy=copy,
                                           subok=subok, **metadata)

    # simple slice, need to copy x0 away from self
    def __getslice__(self, i, j):
        new = super(Array2D, self).__getslice__(i, j).copy()
        new.x0 = float(self.x0.value)
        new.x0 += (i * new.dx)
        return new

    # rebuild getitem to handle complex slicing
    def __getitem__(self, item):
        new = super(Array2D, self).__getitem__(item).copy()
        # if given an int, extract a column
        if isinstance(item, int):
            new = Series(new, unit=self.unit, name=self.name, dx=self.dy,
                         epoch=self.epoch, channel=self.channel, x0=self.y0)
            new.xunit = self.yunit
            return new
        # if given a tuple, extract an element
        elif isinstance(item, tuple) and len(item) == 1:
            return Quantity(new, unit=self.unit)
        # otherwise perform complex slice
        elif isinstance(item, slice):
            if item.start:
                new.x0 += item.start * self.dx
            if item.step:
                new.dx *= item.step
            return new
        else:
            return new

    # -------------------------------------------
    # Series properties

    @property
    def x0(self):
        """X-axis value of first sample
        """
        return self.metadata['x0']

    @x0.setter
    def x0(self, value):
        if isinstance(value, Quantity):
            self.metadata['x0'] = value.to(self.xunit)
        else:
            self.metadata['x0'] = Quantity(value, self.xunit)

    @x0.deleter
    def x0(self):
        del self.metadata['x0']

    @property
    def dx(self):
        """Distance between samples on the x-axis
        """
        return self.metadata['dx']

    @dx.setter
    def dx(self, value):
        if isinstance(value, Quantity):
            self.metadata['dx'] = value.to(self.xunit)
        else:
            self.metadata['dx'] = Quantity(value, self.xunit)

    @dx.deleter
    def dx(self):
        del self.metadata['dx']

    @property
    def span_x(self):
        """Extent of this `Array2D`
        """
        return Segment(self.x0, self.x0 + self.shape[0] * self.dx)

    @property
    def y0(self):
        """X-axis value of first sample
        """
        return self.metadata['y0']

    @y0.setter
    def y0(self, value):
        if isinstance(value, Quantity):
            self.metadata['y0'] = value.to(self.yunit)
        else:
            self.metadata['y0'] = Quantity(value, self.yunit)

    @y0.deleter
    def y0(self):
        del self.metadata['y0']

    @property
    def dy(self):
        """Distance between samples on the x-axis
        """
        return self.metadata['dy']

    @dy.setter
    def dy(self, value):
        if isinstance(value, Quantity):
            self.metadata['dy'] = value.to(self.yunit)
        else:
            self.metadata['dy'] = Quantity(value, self.yunit)

    @dy.deleter
    def dy(self):
        del self.metadata['dy']

    @property
    def span_y(self):
        """Extent of this `Array2D`
        """
        return Segment(self.y0, self.y0 + self.shape[1] * self.dy)

    @property
    def xindex(self):
        """Positions of the data on the x-axis

        :type: `Series`
        """
        try:
            return self._xindex
        except AttributeError:
            if self.logx:
                logdx = (numpy.log10(self.x0.value + self.dx.value) -
                         numpy.log10(self.x0.value))
                logx1 = numpy.log10(self.x0.value) + self.shape[0] * logdx
                self.xindex = numpy.logspace(math.log10(self.x0.value), logx1,
                                             num=self.shape[0])
            else:
                self.xindex = (numpy.arange(self.shape[0]) * self.dx.value +
                               self.x0.value)
            return self.xindex

    @xindex.setter
    def xindex(self, samples):
        if not isinstance(samples, Array):
            fname = inspect.stack()[0][3]
            name = '%s %s' % (self.name, fname)
            samples = Array(samples, unit=self.xunit, name=name,
                            epoch=self.epoch, channel=self.channel)
        self._xindex = samples
        self.x0 = self.xindex[0]
        try:
            self.dx = self.xindex[1] - self.xindex[0]
        except IndexError:
            pass

    @property
    def yindex(self):
        """Positions of the data on the y-axis

        :type: `Series`
        """
        try:
            return self._yindex
        except AttributeError:
            if self.logy:
                logdy = (numpy.log10(self.y0.value + self.dy.value) -
                         numpy.log10(self.y0.value))
                logy1 = numpy.log10(self.y0.value) + self.shape[-1] * logdy
                self.yindex = numpy.logspace(math.log10(self.y0.value), logy1,
                                             num=self.shape[-1])
            else:
                self.yindex = (numpy.arange(self.shape[-1]) * self.dy.value +
                               self.y0.value)
            return self.yindex

    @yindex.setter
    def yindex(self, samples):
        if not isinstance(samples, Array):
            fname = inspect.stack()[0][3]
            name = '%s %s' % (self.name, fname)
            samples = Array(samples, unit=self.yunit, name=name,
                            epoch=self.epoch, channel=self.channel)
        self._yindex = samples
        self.y0 = self.yindex[0]
        try:
            self.dy = self.yindex[1] - self.yindex[0]
        except IndexError:
            pass

    @property
    def logx(self):
        """Boolean telling whether this `Array2D` has a logarithmic
        x-axis scale
        """
        try:
            return self.metadata['logx']
        except KeyError:
            self.logx = False
            return self.logx

    @logx.setter
    def logx(self, val):
        if (val and 'logx' in self.metadata and not self.metdata['logx'] and
                hasattr(self, '_xindex')):
            del self._xindex
        self.metadata['logx'] = bool(val)

    @property
    def logy(self):
        """Boolean telling whether this `Array2D` has a logarithmic
        y-ayis scale
        """
        try:
            return self.metadata['logy']
        except KeyError:
            self.logy = False
            return self.logy

    @logy.setter
    def logy(self, val):
        if (val and 'logy' in self.metadata and not self.metadata['logy'] and
                hasattr(self, '_yindex')):
            del self._index
        self.metadata['logy'] = bool(val)

    # -------------------------------------------
    # Array2D methods

    def resample(self, rate, window=None):
        """Resample this Array2D to a new rate

        Parameters
        ----------
        rate : `float`
            rate to which to resample this `Array2D`
        window : array_like, callable, string, float, or tuple, optional
            specifies the window applied to the signal in the Fourier
            domain.

        Returns
        -------
        Array2D
            a new Array2D with the resampling applied, and the same
            metadata
        """
        if isinstance(rate, Quantity):
            rate = rate.value
        n = self.size * self.dx * rate
        data = signal.resample(self.data, n, window=window)
        new = self.__class__(data, **self.metadata)
        new.dx = 1 / rate
        return new

    # -------------------------------------------
    # numpy.ndarray method modifiers
    # all of these try to return Quantities rather than simple numbers

    def max(self, *args, **kwargs):
        out = super(Array2D, self).max(*args, **kwargs)
        if isinstance(out, Array) and out.shape:
            return Series(out, name='%s max' % self.name, unit=self.unit,
                          x0=out.y0.value, dx=out.dy.value)
        else:
            return out * self.unit
    max.__doc__ = Array.max.__doc__

    def min(self, *args, **kwargs):
        out = super(Array2D, self).min(*args, **kwargs)
        if isinstance(out, Array) and out.shape:
            return Series(out, name='%s min' % self.name, unit=self.unit,
                          x0=out.y0.value, dx=out.dy.value)
        else:
            return out * self.unit
    min.__doc__ = Array.min.__doc__

    def mean(self, *args, **kwargs):
        out = super(Array2D, self).mean(*args, **kwargs)
        if isinstance(out, Array) and out.shape:
            return Series(out, name='%s mean' % self.name, unit=self.unit,
                          x0=out.y0.value, dx=out.dy.value)
        else:
            return out * self.unit
    mean.__doc__ = Array.mean.__doc__

    def median(self, *args, **kwargs):
        out = super(Array2D, self).median(*args, **kwargs)
        if isinstance(out, Array) and out.ndim == 1:
            return Series(out.data, name='%s median' % self.name,
                          unit=self.unit, x0=self.y0.value, dx=self.dy.value)
        else:
            return out * self.unit
    median.__doc__ = Array.median.__doc__



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Common tools for time-axis array.
"""

import warnings
from math import (ceil, floor)

import numpy

from astropy import units
from astropy.time import Time

from .. import version
from ..data import Series
from ..segments import Segment

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version


def is_contiguous(self, other):
    """Check whether other is contiguous with self.
    """
    self.is_compatible(other)
    if abs(float(self.span[1] - other.span[0])) < 1e-5:
        return 1
    elif abs(float(other.span[1] - self.span[0])) < 1e-5:
        return -1
    else:
        return 0


def append(self, other, gap='raise', inplace=True, pad=0.0, resize=True):
    """Connect another `TimeSeries` onto the end of the current one.

    Parameters
    ----------
    other : `TimeSeries`
        the second data set to connect to this one
    gap : `str`, optional, default: ``'raise'``
        action to perform if there's a gap between the other series
        and this one. One of

            - ``'raise'`` - raise an `Exception`
                - ``'ignore'`` - remove gap and join data
            - ``'pad'`` - pad gap with zeros

    inplace : `bool`, optional, default: `True`
        perform operation in-place, modifying current `TimeSeries,
        otherwise copy data and return new `TimeSeries`
    pad : `float`, optional, default: ``0.0``
        value with which to pad discontiguous `TimeSeries`

    Returns
    -------
    series : `TimeSeries`
        time-series containing joined data sets
    """
    # check metadata
    self.is_compatible(other)
    # make copy if needed
    if inplace:
        new = self
    else:
        new = self.copy()
    # fill gap
    if new.is_contiguous(other) != 1:
        if gap == 'pad':
            ngap = (other.span[0] - new.span[1]) // new.dt.value
            if ngap < 1:
                raise ValueError("Cannot append TimeSeries that starts "
                                 "before this one.")
            gapshape = list(new.shape)
            gapshape[0] = int(ngap)
            padding = numpy.ones(gapshape).view(new.__class__) * pad
            padding.epoch = new.span[1]
            padding.dt = new.dt
            padding.unit = new.unit
            new.append(padding, inplace=True, resize=resize)
        elif gap == 'ignore':
            pass
        elif new.span[0] < other.span[0] < new.span[1]:
            raise ValueError("Cannot append overlapping TimeSeries")
        else:
            raise ValueError("Cannot append discontiguous TimeSeries\n"
                             "    TimeSeries 1 span: %s\n"
                             "    TimeSeries 2 span: %s"
                             % (self.span, other.span))
    # check empty other
    if not other.size:
        return new
    # resize first
    if resize:
        s = list(new.shape)
        s[0] = new.shape[0] + other.shape[0]
        try:
            new.resize(s, refcheck=False)
        except ValueError as e:
            if 'resize only works on single-segment arrays' in str(e):
                new = new.copy()
                new.resize(s, refcheck=False)
            else:
                raise
    else:
        new.data[:-other.shape[0]] = new.data[other.shape[0]:]
    new[-other.shape[0]:] = other.data
    try:
        if isinstance(self, Series):
            times = new._index
        else:
            timees = new._xindex
    except AttributeError:
        if not resize:
            new.x0 = new.x0.value + other.shape[0] * new.dx.value
    else:
        if resize:
            new.times.resize(s, refcheck=False)
        else:
            new.times[:-other.shape[0]] = new.times[other.shape[0]:]
        new.times[-other.shape[0]:] = other.times.data
        new.epoch = new.times[0]
    return new


def prepend(self, other, gap='raise', inplace=True, pad=0.0):
    """Connect another `TimeSeries` onto the start of the current one.

    Parameters
    ----------
    other : `TimeSeries`
        the second data set to connect to this one
    gap : `str`, optional, default: ``'raise'``
        action to perform if there's a gap between the other series
        and this one. One of

            - ``'raise'`` - raise an `Exception`
            - ``'ignore'`` - remove gap and join data
            - ``'pad'`` - pad gap with zeros

    inplace : `bool`, optional, default: `True`
        perform operation in-place, modifying current `TimeSeries,
        otherwise copy data and return new `TimeSeries`
    pad : `float`, optional, default: ``0.0``
        value with which to pad discontiguous `TimeSeries`

    Returns
    -------
    series : `TimeSeries`
        time-series containing joined data sets
    """
    # check metadata
    self.is_compatible(other)
    # make copy if needed
    if inplace:
        new = self
    else:
        new = self.copy()
    # fill gap
    if new.is_contiguous(other) != -1:
        if gap == 'pad':
            ngap = (new.span[0]-other.span[1]) // new.dt.value
            if ngap < 1:
                raise ValueError("Cannot prepend TimeSeries that starts "
                                 "after this one.")
            gapshape = list(new.shape)
            gapshape[0] = ngap
            padding = numpy.ones(gapshape).view(new.__class__) * pad
            padding.epoch = other.span[1]
            padding.dt = new.dt
            padding.unit = new.unit
            new.prepend(padding, inplace=True)
        elif gap == 'ignore':
            pass
        elif other.span[0] < new.span[0] < other.span[1]:
            raise ValueError("Cannot prepend overlapping TimeSeries")
        else:
            raise ValueError("Cannot prepend discontiguous TimeSeries")
    # resize first
    N = new.shape[0]
    s = list(new.shape)
    s[0] = new.shape[0] + other.shape[0]
    new.resize(s, refcheck=False)
    new[-N:] = new.data[:N]
    new[:other.shape[0]] = other.data
    return new


def update(self, other, inplace=True):
    """Update this `TimeSeries` by appending new data from an other
    and dropping the same amount of data off the start.

    """
    return self.append(other, inplace=inplace, resize=False)


def crop(self, start=None, end=None, copy=False):
    """Crop this `TimeSeries` to the given GPS ``[start, end)``
    `Segment`.

    Parameters
    ----------
    start : `Time`, `float`
        GPS start time to crop `TimeSeries` at left
    end : `Time`, `float`
        GPS end time to crop `TimeSeries` at right

    Returns
    -------
    timeseries : `TimeSeries`
        A new `TimeSeries` with the same metadata but different GPS
        span

    Notes
    -----
    If either ``start`` or ``end`` are outside of the original
    `TimeSeries` span, warnings will be printed and the limits will
    be restricted to the :attr:`TimeSeries.span`
    """
    # check type
    if isinstance(start, Time):
        start = start.gps
    if isinstance(end, Time):
        end = end.gps
    # pin early starts to time-series start
    if start == self.span[0]:
        start = None
    elif start is not None and start < self.span[0]:
        warnings.warn('TimeSeries.crop given GPS start earlier than '
                      'start time of the input TimeSeries. Crop will '
                      'begin when the TimeSeries actually starts.')
        start = None
    # pin late ends to time-series end
    if end == self.span[1]:
        end = None
    if start is not None and end > self.span[1]:
        warnings.warn('TimeSeries.crop given GPS end later than '
                      'end time of the input TimeSeries. Crop will '
                      'end when the TimeSeries actually ends.')
        end = None
    # find start index
    if start is None:
        idx0 = None
    else:
        idx0 = float(start - self.span[0]) // self.dt.value
    # find end index
    if end is None:
        idx1 = None
    else:
        idx1 = float(end - self.span[0]) // self.dt.value
        if idx1 >= self.size:
            idx1 = None
    # crop
    if copy:
        return self[idx0:idx1].copy()
    else:
        return self[idx0:idx1]



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Module providing classes for reading, writing and manipulating
time-series and spectrum data.
"""

from .. import version
__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__version__ = version.version

from glue.lal import (Cache, CacheEntry)

from .array import *
from .array2d import *
from .series import *

# define custom time-series units
try:
    from astropy import (__version__ as astropyversion, units)
except ImportError:
    pass
else:
    from distutils.version import LooseVersion
    if LooseVersion(astropyversion) < LooseVersion('0.3'):
        units.def_unit(['counts'], represents=units.Unit('count'),
                       register=True)
        units.def_unit(['strain'], represents=units.dimensionless_unscaled,
                       register=True)
        units.def_unit(['coherence'], represents=units.dimensionless_unscaled,
                       register=True)
    else:
        units.add_enabled_units([
            units.def_unit(['counts'], units.Unit('count')),
            units.def_unit(['coherence'], units.dimensionless_unscaled),
            units.def_unit(['strain'], units.dimensionless_unscaled),
            ])

__all__ = ['Array', 'Array2D', 'Series', 'Cache', 'CacheEntry']




# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read data from gravitational-wave frames using frameCPP.

Direct access to the frameCPP library is the easiest way to read multiple
channels from a single frame file in one go.
"""

from __future__ import division

from astropy.io import registry

from glue.lal import (Cache, CacheEntry)

from .identify import register_identifier
from .... import version
from ....utils import (gprint, with_import)
from ... import (TimeSeries, TimeSeriesDict, StateVector, StateVectorDict)

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version


@with_import('frameCPP')
def read_timeseriesdict(source, channels, start=None, end=None, type=None,
                        resample=None, verbose=False, _SeriesClass=TimeSeries):
    """Read the data for a list of channels from a GWF data source.

    Parameters
    ----------
    source : `str`, :class:`glue.lal.Cache`, `list`
        data source object, one of:

        - `str` : frame file path
        - :class:`glue.lal.Cache`, `list` : contiguous list of frame paths

    channels : `list`
        list of channel names (or `Channel` objects) to read from frame.
    start : `Time`, :lalsuite:`LIGOTimeGPS`, optional
        start GPS time of desired data.
    end : `Time`, :lalsuite:`LIGOTimeGPS`, optional
        end GPS time of desired data.
    type : `str`
        type of channel, one of ``'adc'`` or ``'proc'``.
    resample : `float`, optional
        rate of samples per second at which to resample input TimeSeries.
    verbose : `bool`, optional
        print verbose output.

    Returns
    -------
    dict : :class:`~gwpy.timeseries.core.TimeSeriesDict`
        dict of (channel, `TimeSeries`) data pairs

    Notes
    -----
    If reading from a list, or cache, or framefiles, the frames contained
    must be contiguous and sorted in chronological order for this function
    to return without exception.

    Raises
    ------
    ValueError
        if reading from an unsorted, or discontiguous cache of files
    """
    # parse input source
    if isinstance(source, file):
        filelist = [source.name]
    elif isinstance(source, (unicode, str)):
        filelist = source.split(',')
    elif isinstance(source, CacheEntry):
        filelist = [source]
    elif isinstance(source, Cache):
        source.sort(key=lambda e: e.segment[0])
        filelist = source
    else:
        filelist = list(source)
    # parse resampling
    if isinstance(resample, int):
        resample = dict((channel, resample) for channel in channels)
    elif isinstance(resample, (tuple, list)):
        resample = dict(zip(channels, resample))
    elif resample is None:
        resample = {}
    elif not isinstance(resample, dict):
        raise ValueError("Cannot parse resample request, please review "
                         "documentation for that argument")
    # read each individually and append
    N = len(filelist)
    if verbose:
        if not isinstance(verbose, (unicode, str)):
            verbose = ''
        gprint("%sReading %d channels from frames... 0/%d (0.00%%)\r"
               % (verbose, len(channels), N), end='')
    out = TimeSeriesDict()
    for i,fp in enumerate(filelist):
        # read frame
        new = _read_frame(fp, channels, type=type, verbose=verbose,
                          _SeriesClass=_SeriesClass)
        # store
        out.append(new)
        if verbose is not False:
            gprint("%sReading %d channels from frames... %d/%d (%.1f%%)\r"
                   % (verbose, len(channels), i+1, N, (i+1)/N * 100), end='')
    if verbose is not False:
        gprint("%sReading %d channels from frames... %d/%d (100.0%%)"
               % (verbose, len(channels), N, N))
    # finalise
    for channel, ts in out.iteritems():
        # resample data
        if resample is not None and channel in resample:
            out[channel] = out[channel].resample(resample[channel])
        # crop data
        if start is not None or end is not None:
            out[channel] = out[channel].crop(start=start, end=end)
    return out


def _read_frame(framefile, channels, type=None, verbose=False,
                _SeriesClass=TimeSeries):
    """Internal function to read data from a single frame.

    All users should be using the wrapper `read_timeseriesdict`.

    Parameters
    ----------
    framefile : `str`, :class:`~glue.lal.CacheEntry`
        path to GWF-format frame file on disk.
    channels : `list`
        list of channels to read.
    type : `str`, optional
        channel data type to read, one of: ``'adc'``, ``'proc'``.
    verbose : `bool`, optional
        print verbose output, optional, default: `False`
    _SeriesClass : `type`, optional
        class object to use as the data holder for a single channel,
        default is :class:`~gwpy.timeseries.core.TimeSeries`

    Returns
    -------
    dict : :class:`~gwpy.timeseries.core.TimeSeriesDict`
        dict of (channel, `TimeSeries`) data pairs
    """
    if isinstance(channels, (unicode, str)):
        channels = channels.split(',')

    # open file
    if isinstance(framefile, CacheEntry):
        fp = framefile.path
    else:
        fp = framefile
    stream = frameCPP.IFrameFStream(fp)

    # interpolate frame epochs from CacheEntry
    # FIXME: update when new frameCPP is released
    nframe = 0 # int(stream.GetNumberOfFrames())
    if isinstance(framefile, CacheEntry) and nframe == 1:
        epochs = [framefile.segment[0]]
    else:
        epochs = None

    # load table of contents if needed
    if epochs is None or not type:
        toc = stream.GetTOC()
    # get list of frame epochs
    if epochs is None:
        epochs = toc.GTimeS
    # work out channel types
    if type:
        ctype = dict((str(channel), type) for channel in channels)
    else:
        try:
            adcs = toc.GetADC().keys()
        except AttributeError:
            adcs = toc.GetADC()
        try:
            procs = toc.GetProc().keys()
        except AttributeError:
            procs = toc.GetProc()
        ctype = {}
        for channel in channels:
            name = str(channel)
            if name in adcs:
                ctype[name] = 'adc'
            elif name in procs:
                ctype[name] = 'proc'
            else:
                raise ValueError("Channel %s not found in frame table of "
                                 "contents" % name)

    # set output
    out = TimeSeriesDict()
    for channel in channels:
        name = str(channel)
        read_ = getattr(stream, 'ReadFr%sData' % ctype[name].title())
        ts = None
        i = 0
        while True:
            try:
                data = read_(i, name)
            except IndexError:
                break
            offset = data.GetTimeOffset()
            thisepoch = epochs[i] + offset
            for vect in data.data:
                arr = vect.GetDataArray()
                dx = vect.GetDim(0).dx
                if ts is None:
                    unit = vect.GetUnitY()
                    ts = _SeriesClass(arr, epoch=thisepoch, dx=dx, name=name,
                                      channel=channel, unit=unit, copy=True)
                else:
                    ts.append(arr)
            i += 1
        if ts is None:
            raise ValueError("Channel '%s' not found in frame '%s'"
                             % (str(channel), fp))
        else:
            out[channel] = ts

    return out


@with_import('frameCPP')
def read_timeseries(source, channel, **kwargs):
    """Read a `TimeSeries` of data from a gravitational-wave frame file

    Parameters
    ----------
    source : `str`, :class:`glue.lal.Cache`, `list`
        data source object, one of:

        - `str` : frame file path
        - :class:`glue.lal.Cache`, `list` : contiguous list of frame paths

    channel : :class:`~gwpy.detector.channel.Channel`, `str`
        data channel to read from frames

    See Also
    --------
    :func:`~gwpy.io.gwf.framecpp.read_timeseriesdict`
        for documentation on keyword arguments

    Returns
    -------
    data : :class:`~gwpy.timeseries.core.TimeSeries`
        a new `TimeSeries` containing the data read from disk
    """
    return read_timeseriesdict(source, [channel], **kwargs)[channel]


@with_import('frameCPP')
def read_statevectordict(source, channels, bitss=[], **kwargs):
    """Read a `StateVectorDict` of data from a gravitational-wave
    frame file.
    """
    kwargs.setdefault('_SeriesClass', StateVector)
    svd = StateVectorDict(read_timeseriesdict(source, channels, **kwargs))
    for (channel, bits) in zip(channels, bitss):
        svd[channel].bits = bits
    return svd


@with_import('frameCPP')
def read_statevector(source, channel, bits=[], **kwargs):
    """Read a `StateVector` of data from a gravitational-wave frame file

    Parameters
    ----------
    source : `str`, :class:`glue.lal.Cache`, `list`
        data source object, one of:

        - `str` : frame file path
        - :class:`glue.lal.Cache`, `list` : contiguous list of frame paths

    channel : :class:`~gwpy.detector.channel.Channel`, `str`
        data channel to read from frames
    bits : `list`
        ordered list of bit identifiers (names)

    See Also
    --------
    :func:`~gwpy.io.gwf.framecpp.read_timeseriesdict`
        for documentation on keyword arguments

    Returns
    -------
    data : :class:`~gwpy.timeseries.core.StateVector`
        a new `StateVector` containing the data read from disk
    """
    kwargs.setdefault('_SeriesClass', StateVector)
    sv = read_timeseries(source, channel, **kwargs)
    sv.bits = bits
    return sv


# register gwf reader first
try:
    import frameCPP
except ImportError:
    pass
else:
    try:
        register_identifier('gwf')
    except Exception as e:
        if not str(e).startswith('Identifier for format'):
            raise
    registry.register_reader('gwf', TimeSeriesDict, read_timeseriesdict,
                             force=True)
    registry.register_reader('gwf', StateVectorDict, read_statevectordict,
                             force=True)
    try:
        registry.register_reader('gwf', TimeSeries, read_timeseries, force=True)
    except:
        pass
    else:
        registry.register_reader('gwf', StateVector, read_statevector,
                                 force=True)

# register framecpp
registry.register_reader('framecpp', TimeSeriesDict, read_timeseriesdict)
registry.register_reader('framecpp', TimeSeries, read_timeseries)
registry.register_reader('framecpp', StateVectorDict, read_statevectordict)
registry.register_reader('framecpp', StateVector, read_statevector)




# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Extension of the simple Plot class for displaying segment objects
"""

import operator
import re

import numpy

from matplotlib.ticker import (Formatter, MultipleLocator, NullLocator)
from matplotlib.projections import register_projection
from matplotlib.collections import PatchCollection
from matplotlib.patches import Rectangle
try:
    from mpl_toolkits.axes_grid1 import make_axes_locatable
except ImportError:
    from mpl_toolkits.axes_grid import make_axes_locatable

from .. import version
from . import rcParams
from .timeseries import (TimeSeriesPlot, TimeSeriesAxes)
from .decorators import auto_refresh
from ..segments import *
from ..timeseries import StateVector

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version


class SegmentAxes(TimeSeriesAxes):
    """Custom `Axes` for a :class:`~gwpy.plotter.SegmentPlot`.

    This `SegmentAxes` provides custom methods for displaying any of

    - :class:`~gwpy.segments.flag.DataQualityFlag`
    - :class:`~gwpy.segments.segments.Segment`
    - :class:`~gwpy.segments.segments.SegmentList`
    - :class:`~gwpy.segments.segments.SegmentListDict`
    """
    name = 'segments'

    def __init__(self, *args, **kwargs):
        super(SegmentAxes, self).__init__(*args, **kwargs)
        self.yaxis.set_major_locator(MultipleLocator())

    def plot(self, *args, **kwargs):
        """Plot data onto these axes

        Parameters
        ----------
        args
            a single instance of

                - :class:`~gwpy.segments.flag.DataQualityFlag`
                - :class:`~gwpy.segments.segments.Segment`
                - :class:`~gwpy.segments.segments.SegmentList`
                - :class:`~gwpy.segments.segments.SegmentListDict`

        kwargs
            keyword arguments applicable to :meth:`~matplotib.axes.Axes.plot`

        Returns
        -------
        Line2D
            the :class:`~matplotlib.lines.Line2D` for this line layer

        See Also
        --------
        :meth:`matplotlib.axes.Axes.plot`
            for a full description of acceptable ``*args` and ``**kwargs``
        """
        lim = len(self.collections)
        out = []
        args = list(args)
        while len(args):
            if isinstance(args[0], DataQualityFlag):
                out.append(self.plot_dqflag(args[0], **kwargs))
                args.pop(0)
                continue
            elif isinstance(args[0], SegmentListDict):
                out.extend(self.plot_segmentlistdict(args[0], **kwargs))
                args.pop(0)
                continue
            elif isinstance(args[0], SegmentList):
                out.append(self.plot_segmentlist(args[0], **kwargs))
                args.pop(0)
                continue
            elif isinstance(args[0], Segment):
                raise ValueError("Input must be DataQualityFlag, "
                                 "SegmentListDict, or SegmentList")
            break
        if len(args):
            out.append(super(SegmentAxes, self).plot(*args, **kwargs))
        if not lim:
            self.set_ylim(-0.1, len(self.collections) + 0.1)
        return out

    @auto_refresh
    def plot_dqflag(self, flag, y=None, valid='x', add_label=True, **kwargs):
        """Plot a :class:`~gwpy.segments.flag.DataQualityFlag`
        onto these axes

        Parameters
        ----------
        flag : :class:`~gwpy.segments.flag.DataQualityFlag`
            data-quality flag to display
        y : `float`, optional
            y-axis value for new segments
        height : `float`, optional, default: 0.8
            height for each segment block
        valid : `str`, `dict`, `None`, default: '/'
            display `valid` segments with the given hatching, or give a
            dict of keyword arguments to pass to
            :meth:`~SegmentAxes.plot_segmentlist`, or `None` to hide.
        add_label : `bool`, default: `True`
            add a label to the y-axis for this `DataQualityFlag`
        **kwargs
            any other keyword arguments acceptable for
            :class:`~matplotlib.patches.Rectangle`

        Returns
        -------
        collection : :class:`~matplotlib.patches.PatchCollection`
            list of :class:`~matplotlib.patches.Rectangle` patches
        """
        # get y axis position
        if y is None:
            y = len(self.collections)
        # get flag name
        name = kwargs.pop('label', flag.texname)

        # get epoch
        try:
            if not self.epoch:
                self.set_epoch(flag.valid[0][0])
            else:
                self.set_epoch(min(self.epoch, flag.valid[0][0]))
        except IndexError:
            pass
        # make valid collection
        if valid is not None:
            if isinstance(valid, dict):
                vkwargs = valid
            else:
                vkwargs = kwargs.copy()
                vkwargs.pop('label', None)
                vkwargs['fill'] = False
                vkwargs['hatch'] = valid
            vkwargs['collection'] = False
            vkwargs['zorder'] = -1000
            self.plot_segmentlist(flag.valid, y=y, label=None, **vkwargs)
        # make active collection
        collection = self.plot_segmentlist(flag.active, y=y, label=name,
                                           **kwargs)
        if len(self.collections) == 1:
            if len(flag.valid):
                self.set_xlim(*map(float, flag.extent))
            self.autoscale(axis='y')
        # add label
        if add_label:
            self.label_segments(y, name, inset=(add_label == 'inset'))
        return collection

    @auto_refresh
    def plot_segmentlist(self, segmentlist, y=None, collection=True,
                         label=None, add_label=True, **kwargs):
        """Plot a :class:`~gwpy.segments.segments.SegmentList` onto
        these axes

        Parameters
        ----------
        segmentlist : :class:`~gwpy.segments.segments.SegmentList`
            list of segments to display
        y : `float`, optional
            y-axis value for new segments
        collection : `bool`, default: `True`
            add all patches as a
            :class:`~matplotlib.collections.PatchCollection`, doesn't seem
            to work for hatched rectangles
        label : `str`, optional
            custom descriptive name to print as y-axis tick label
        add_label : `bool`, `str`, optional
            if `True` print label on y-axis, if ``'inset'`` print inside
            axes, otherwise ignore.
        **kwargs
            any other keyword arguments acceptable for
            :class:`~matplotlib.patches.Rectangle`

        Returns
        -------
        collection : :class:`~matplotlib.patches.PatchCollection`
            list of :class:`~matplotlib.patches.Rectangle` patches
        """
        if y is None:
            y = len(self.collections)
        patches = []
        for seg in segmentlist:
            patches.append(self.build_segment(seg, y, **kwargs))
        try:
            if not self.epoch:
                self.set_epoch(segmentlist[0][0])
            else:
                self.set_epoch(min(self.epoch, segmentlist[0][0]))
        except IndexError:
            pass
        if label and add_label:
            self.label_segments(y, label, inset=(add_label == 'inset'))
        if collection:
            return self.add_collection(PatchCollection(patches,
                                                       len(patches) != 0))
        else:
            out = []
            for p in patches:
                out.append(self.add_patch(p))
            return out

    @auto_refresh
    def plot_segmentlistdict(self, segmentlistdict, y=None, dy=1, **kwargs):
        """Plot a :class:`~gwpy.segments.segments.SegmentListDict` onto
        these axes

        Parameters
        ----------
        segmentlistdict : :class:`~gwpy.segments.segments.SegmentListDict`
            (name, :class:`~gwpy.segments.segments.SegmentList`) dict
        y : `float`, optional
            starting y-axis value for new segmentlists
        **kwargs
            any other keyword arguments acceptable for
            :class:`~matplotlib.patches.Rectangle`

        Returns
        -------
        collections : `list`
            list of :class:`~matplotlib.patches.PatchCollection` sets for
            each segmentlist
        """
        if y is None:
            y = len(self.collections)
        collections = []
        for name, segmentlist in segmentlistdict.iteritems():
            collections.append(self.plot_segmentlist(segmentlist, y=y,
                                                     label=name, **kwargs))
            y += dy
        return collections

    @staticmethod
    def build_segment(segment, y, height=.8, valign='center', **kwargs):
        """Build a :class:`~matplotlib.patches.Rectangle` to display
        a single :class:`~gwpy.segments.segments.Segment`

        Parameters
        ----------
        segment : :class:`~gwpy.segments.segments.Segment`
            [start, stop) GPS segment
        y : `float`
            y-axis position for segment
        height : `float`, optional, default: 1
            height (in y-axis units) for segment
        valign : `str`
            alignment of segment on y-axis value:
            `top`, `center`, or `bottom`
        **kwargs
            any other keyword arguments acceptable for
            :class:`~matplotlib.patches.Rectangle`

        Returns
        -------
        box : `~matplotlib.patches.Rectangle`
            rectangle patch for segment display
        """
        if valign.lower() == 'center':
            y0 = y - height/2.
        elif valign.lower() == 'top':
            y0 = y - height
        else:
            raise ValueError("valign must be one of 'top', 'center', or "
                             "'bottom'")
        return Rectangle((segment[0], y0), width=abs(segment), height=height,
                         **kwargs)

    @auto_refresh
    def label_segments(self, y, label, inset=False, **insetparams):
        """Replace the default Y-axis label with a custom string at a
        given Y-axis position ``y``.

        Parameters
        ----------
        y : `int`
            Y-axis position to modify
        label : `str`
            custom text to insert
        inset : `bool`, optional, default: `False`
            place the label inside the axes, rather than outside
            (default)
        **insetparams
            other keyword arguments for the inset box
        """
        if label is not None:
            label = re.sub('r\\+_+', '\_', label)
        # find existing label
        # set formatter
        formatter = self.yaxis.get_major_formatter()
        if not isinstance(formatter, SegmentFormatter):
            formatter = SegmentFormatter()
            self.yaxis.set_major_formatter(formatter)

        # hide existing label and add inset text
        if inset:
            xlim = self.get_xlim()
            x = xlim[0] + (xlim[1]-xlim[0]) * 0.01
            formatter.flags[y] = ""
            insetparams.setdefault('fontsize', rcParams['axes.labelsize'])
            insetparams.setdefault('horizontalalignment', 'left')
            insetparams.setdefault('verticalalignment', 'center')
            insetparams.setdefault('backgroundcolor', 'white')
            insetparams.setdefault('transform', self.transData)
            insetparams.setdefault('bbox',
                                   {'alpha': 0.5, 'facecolor': 'white',
                                    'edgecolor': 'none'})
            t = self.text(x, y - 0.1, label or '', **insetparams)
            t._is_segment_label = True
            return t
        else:
            formatter.flags[y] = label or ''
            return

    def set_xlim(self, *args, **kwargs):
        out = super(SegmentAxes, self).set_xlim(*args, **kwargs)
        _xlim = self.get_xlim()
        try:
            texts = self.texts
        except AttributeError:
            pass
        else:
            for t in texts:
                if hasattr(t, '_is_segment_label') and t._is_segment_label:
                    t.set_x(_xlim[0] + (_xlim[1] - _xlim[0]) * 0.01)
        return out
    set_xlim.__doc__ = TimeSeriesAxes.set_xlim.__doc__

register_projection(SegmentAxes)


class SegmentPlot(TimeSeriesPlot):
    """`Figure` for displaying a :class:`~gwpy.segments.flag.DataQualityFlag`.

    Parameters
    ----------
    *flags : `DataQualityFlag`
        any number of :class:`~gwpy.segments.flag.DataQualityFlag` to
        display on the plot
    **kwargs
        other keyword arguments as applicable for the
        :class:`~gwpy.plotter.Plot`
    """
    _DefaultAxesClass = SegmentAxes

    def __init__(self, *flags, **kwargs):
        """Initialise a new SegmentPlot
        """
        # separate kwargs into figure args and plotting args
        figargs = {}
        if 'figsize' in kwargs:
            figargs['figsize'] = kwargs.pop('figsize')
        sep = kwargs.pop('sep', False)
        epoch = kwargs.pop('epoch', None)

        # generate figure
        super(SegmentPlot, self).__init__(**figargs)
        # plot data
        for flag in flags:
            self.add_dataqualityflag(flag,
                                     projection=self._DefaultAxesClass.name,
                                     newax=sep, **kwargs)

        # set epoch
        if len(flags):
            span = reduce(operator.or_, [f.valid for f in flags]).extent()
            if not epoch:
                epoch = span[0]
            for ax in self.axes:
                ax.set_epoch(epoch)
                ax.set_xlim(*map(float, span))
            for ax in self.axes[:-1]:
                ax.set_xlabel("")

        if sep:
            for ax in self.axes:
                ax.set_ylim(-0.5, 0.5)
                ax.grid(b=False, which='both', axis='y')
        elif len(flags):
            ax.set_ylim(-0.5, len(flags)-0.5)
            ax.grid(b=False, which='both', axis='y')

    def add_dataqualityflag(self, flag, **kwargs):
        super(SegmentPlot, self).add_dataqualityflag(flag, **kwargs)
        if self.epoch is None:
            try:
                self.set_epoch(flag.valid[0][0])
            except IndexError:
                pass
    add_dataqualityflag.__doc__ = TimeSeriesPlot.add_dataqualityflag.__doc__

    def add_bitmask(self, mask, ax=None, width=0.2, pad=0.1,
                    visible=True, axes_class=SegmentAxes, topdown=False,
                    **plotargs):
        """Display a state-word bitmask on a new set of Axes.
        """
        # find default axes
        if ax is None:
            ax = self.axes[-1]

        # get new segment axes
        divider = make_axes_locatable(ax)
        max = divider.new_horizontal(size=width, pad=pad,
                                     axes_class=axes_class)
        max.set_xscale('gps')
        max.xaxis.set_major_locator(NullLocator())
        max.xaxis.set_minor_locator(NullLocator())
        max.yaxis.set_minor_locator(NullLocator())
        if visible:
            self.add_axes(max)
        else:
            return

        # format mask as a binary string and work out how many bits to set
        if isinstance(mask, int):
            mask = bin(mask)
        elif isinstance(mask, (unicode, str)) and 'x' in mask:
            mask = bin(int(mask, 16))
        maskint = int(mask, 2)
        if topdown:
            bits = list(range(len(mask.split('b', 1)[1])))[::-1]
        else:
            bits = list(range(len(mask.split('b', 1)[1])))

        # loop over bits
        plotargs.setdefault('facecolor', 'green')
        plotargs.setdefault('edgecolor', 'black')
        s = Segment(0, 1)
        for bit in bits:
            if maskint >> bit & 1:
                sl = SegmentList([s])
            else:
                sl = SegmentList()
            max.plot(sl, **plotargs)
        max.set_title('Bitmask')
        max.set_xlim(0, 1)
        max.set_xticks([])
        max.yaxis.set_ticklabels([])
        max.set_xlabel('')
        max.set_ylim(*ax.get_ylim())

        return max

class SegmentFormatter(Formatter):
    """Custom tick formatter for y-axis flag names
    """
    def __init__(self, flags={}):
        self.flags = flags

    def __call__(self, t, pos=None):
        if t in self.flags:
            return self.flags[t]
        else:
            return t



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""A collection of average power spectral density calculation routines

Average-spectrum calculation routines are available for the following methods

    - :func:`Bartlett <bartlett>`
    - :func:`Welch <welch>`
    - :func:`Median-mean <median_mean>`
    - :func:`Median <median>`

Each of these methods utilises an existing method provided by the
LIGO Algorithm Library, wrapped into python as part of the `lal.spectrum`
module.
"""

import numpy
from matplotlib import mlab
from scipy import signal

from astropy import units

from .core import Spectrum
from ..window import Window

from .. import version
__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__version__ = version.version

__all__ = ['bartlett', 'welch', 'median_mean', 'median', 'spectrogram']

# cache windows internally
LAL_WINDOWS = {}
LAL_FFTPLANS = {}
LAL_FFTPLAN_LEVEL = 1


def bartlett(timeseries, segmentlength, window=None, plan=None):
    """Calculate the power spectral density of the given `TimeSeries`
    using the Bartlett average method.

    This method divides the data into chunks of length `segmentlength`,
    a periodogram calculated for each, and the bin-by-bin mean returned.

    Parameters
    ----------
    timeseries: `TimeSeries`
        input `TimeSeries` data
    segmentlength : `int`
        number of samples in each average
    plan : :lalsuite:`REAL8FFTPlan`, optional
        LAL FFT plan to use when generating average spectrum

    Returns
    -------
    Spectrum
        Bartlett-averaged `Spectrum`
    """
    return welch(timeseries, segmentlength, segmentlength, window=window,
                 plan=None)


def welch(timeseries, segmentlength, overlap, window=None, plan=None):
    """Calculate the power spectral density of the given `TimeSeries`
    using the Welch average method.

    For more details see :lalsuite:`XLALREAL8AverageSpectrumWelch`.

    Parameters
    ----------
    timeseries : `TimeSeries`
        input `TimeSeries` data
    method : `str`
        average method
    segmentlength : `int`
        number of samples in single average
    overlap : `int`
        number of samples between averages
    window : `~gwpy.window.Window`, optional
        window function to apply to timeseries prior to FFT
    plan : :lalsuite:`REAL8FFTPlan`, optional
        LAL FFT plan to use when generating average spectrum

    Returns
    -------
    Spectrum
        Welch-averaged `Spectrum`
    """
    try:
        return lal_psd(timeseries, 'welch', segmentlength, overlap,
                       window=window, plan=None)
    except ImportError:
        return scipy_psd(timeseries, 'welch', segmentlength, overlap,
                         window=window)


def median_mean(timeseries, segmentlength, overlap, window=None, plan=None):
    """Calculate the power spectral density of the given `TimeSeries`
    using the median-mean average method.

    For more details see :lalsuite:`XLALREAL8AverageSpectrumMedianMean`.

    Parameters
    ----------
    timeseries : `TimeSeries`
        input `TimeSeries` data
    segmentlength : `int`
        number of samples in single average
    overlap : `int`
        number of samples between averages
    window : `~gwpy.window.Window`, optional
        window function to apply to timeseries prior to FFT
    plan : :lalsuite:`REAL8FFTPlan`, optional
        LAL FFT plan to use when generating average spectrum

    Returns
    -------
    Spectrum
        median-mean-averaged `Spectrum`
    """
    return lal_psd(timeseries, 'medianmean', segmentlength, overlap,
                    window=window, plan=None)


def median(timeseries, segmentlength, overlap, window=None, plan=None):
    """Calculate the power spectral density of the given `TimeSeries`
    using the median-mean average method.

    For more details see :lalsuite:`XLALREAL8AverageSpectrumMean`.

    Parameters
    ----------
    timeseries : `TimeSeries`
        input `TimeSeries` data
    segmentlength : `int`
        number of samples in single average
    overlap : `int`
        number of samples between averages
    window : `~gwpy.window.Window`, optional
        window function to apply to timeseries prior to FFT
    plan : :lalsuite:`REAL8FFTPlan`, optional
        LAL FFT plan to use when generating average spectrum

    Returns
    -------
    Spectrum
        median-mean-averaged `Spectrum`
    """
    return lal_psd(timeseries, 'medianmean', segmentlength, overlap,
                    window=window, plan=None)


def lal_psd(timeseries, method, segmentlength, overlap, window=None, plan=None):
    """Internal wrapper to the `lal.spectrum.psd` function

    This function handles the conversion between GWpy `TimeSeries` and
    XLAL ``TimeSeries``, (e.g. :lalsuite:`XLALREAL8TimeSeries`).

    Parameters
    ----------
    timeseries : `TimeSeries`
        input `TimeSeries` data
    method : `str`
        average method
    segmentlength : `int`
        number of samples in single average
    overlap : `int`
        number of samples between averages
    window : `~gwpy.window.Window`, optional
        window function to apply to timeseries prior to FFT
    plan : :lalsuite:`REAL8FFTPlan`, optional
        LAL FFT plan to use when generating average spectrum

    Returns
    -------
    Spectrum
        average power `Spectrum`
    """
    try:
        from lal.spectrum import averagespectrum as lalspectrum
    except ImportError as e:
        raise ImportError('%s. Try using gwpy.spectrum.scipy_psd instead'
                          % str(e))
    else:
        from lal import (lal, utils as lalutils)
    if isinstance(segmentlength, units.Quantity):
        segmentlength = segmentlength.value
    if isinstance(overlap, units.Quantity):
        overlap = overlap.value
    lalts = timeseries.to_lal()
    stype = lalutils.dtype(lalts)
    # get cached window
    if window is None:
        window = generate_lal_window(segmentlength, dtype=timeseries.dtype)
    elif isinstance(window, Window):
        window = window.to_lal()
    # get cached FFT plan
    if plan is None:
        plan = generate_lal_fft_plan(segmentlength, dtype=timeseries.dtype)
    # generate average spectrum
    lalfs = lalspectrum._psd(method, lalts, segmentlength, overlap,
                             window=window, plan=plan)
    # format and return
    spec = Spectrum.from_lal(lalfs)
    spec.channel = timeseries.channel
    if timeseries.unit:
        spec.unit = timeseries.unit ** 2 / units.Hertz
    else:
        spec.unit = 1 / units.Hertz
    return spec


def scipy_psd(timeseries, method, segmentlength, overlap, window=('kaiser', 24)):
    """Internal wrapper to the `lal.spectrum.psd` function

    This function handles the conversion between GWpy `TimeSeries` and
    XLAL ``TimeSeries``, (e.g. :lalsuite:`XLALREAL8TimeSeries`).

    Parameters
    ----------
    timeseries : `TimeSeries`
        input `TimeSeries` data
    method : `str`
        average method
    segmentlength : `int`
        number of samples in single average
    overlap : `int`
        number of samples between averages
    window : `~gwpy.window.Window`, optional
        window function to apply to timeseries prior to FFT

    Returns
    -------
    Spectrum
        average power `Spectrum`
    """
    methods = ['welch', 'bartlett']
    if method.lower() not in methods:
        raise ValueError("'method' must be one of: '%s'" % "','".join(methods))
    if isinstance(segmentlength, units.Quantity):
        segmentlength = segmentlength.value
    if isinstance(overlap, units.Quantity):
        overlap = overlap.value
    f, psd_ = signal.welch(timeseries.data, fs=timeseries.sample_rate.value,
                           window=window, nperseg=segmentlength,
                           noverlap=(segmentlength-overlap))
    spec = psd_.view(Spectrum)
    spec.name = timeseries.name
    spec.epoch = timeseries.epoch
    spec.channel = timeseries.channel
    spec.f0 = f[0]
    spec.df = f[1]-f[0]
    if timeseries.unit:
        spec.unit = timeseries.unit ** 2 / units.Hertz
    else:
        spec.unit = 1 / units.Hertz
    return spec


def generate_lal_fft_plan(length, level=None,
                          dtype=numpy.dtype(numpy.float64)):
    """Build a plan to use when performing a Fourier transform using
    the LIGO Algorithm Library routines.

    Parameters
    ----------
    length : `int`
        number of samples to plan for in each FFT.
    level : `int`, optional
        amount of work to do when planning the FFT, default set by
        LAL_FFTPLAN_LEVEL module variable.
    dtype : :class:`numpy.dtype`
        numeric type of data to plan for, default `numpy.dtype(numpy.float64)`

    Returns
    -------
    :lalsuite:`REAL8FFTPlan`
        FFT plan of the relevant data type
    """
    from lal import (lal, utils as lalutils)
    global LAL_FFTPLANS
    ltype = lalutils.LAL_TYPE_FROM_NUMPY[dtype.type]
    laltypestr = lalutils.LAL_TYPE_STR[ltype]
    try:
        plan = LAL_FFTPLANS[(length, laltypestr)]
    except KeyError:
        create = getattr(lal, 'CreateForward%sFFTPlan' % laltypestr)
        if level is None:
           level = LAL_FFTPLAN_LEVEL
        plan = LAL_FFTPLANS[(length, laltypestr)] = create(length, level)
    return plan


def generate_lal_window(length, type=('kaiser', 24),
                        dtype=numpy.dtype(numpy.float64)):
    """Generate a time-domain window for use in a Fourier transform using
    the LIGO Algorithm Library routines.

    Parameters
    ----------
    length : `int`
        length of window in samples.
    type : `str`, `tuple`
        name of window to generate, default: ``('kaiser', 24)``. Give
        `str` for simple windows, or tuple of ``(name, *args)`` for
         complicated windows
    dtype : :class:`numpy.dtype`
        numeric type of window, default `numpy.dtype(numpy.float64)`

    Returns
    -------
    `window` : XLAL window
        time-domain window to use for FFT
    """
    from lal import (lal, utils as lalutils)
    global LAL_WINDOWS
    ltype = lalutils.LAL_TYPE_FROM_NUMPY[dtype.type]
    laltypestr = lalutils.LAL_TYPE_STR[ltype]
    wtype = isinstance(type, (list, tuple)) and type[0] or str(type)
    try:
        window = LAL_WINDOWS[(length, wtype.lower(), laltypestr)]
    except KeyError:
        if isinstance(type, (list, tuple)):
            try:
                args = type[1:]
            except IndexError:
                args = []
        else:
            args = []
        wtype = wtype.islower() and wtype.title() or wtype
        create = getattr(lal, 'Create%s%sWindow' % (wtype, laltypestr))
        window = LAL_WINDOWS[(length, wtype.lower(), laltypestr)] = create(
                                                                 length, *args)
    return window



#!/usr/bin/env python

# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""GWpy Example: plotting a time-series

Problem
-------

I would like to study the gravitational wave strain time-series around the time of an interesting simulated signal during the last science run (S6).

These data are public, so we can load them directly from the web.
"""

from gwpy import version
__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__version__ = version.version

from urllib2 import urlopen

from numpy import asarray

from gwpy.timeseries import TimeSeries

# download data
data = urlopen('http://www.ligo.org/science/GW100916/L-strain_hp30-968654552-10.txt').read()

# parse as floats and read as a TimeSeries
ts = TimeSeries(asarray(data.splitlines(), dtype=float), epoch=968654552, sample_rate=16384, unit='strain')

# make a plot
plot = ts.plot()
plot.set_title('LIGO Livingston Observatory data for GW100916')
plot.set_ylabel('Gravitational-wave strain amplitude')

if __name__ == '__main__':
    try:
        outfile = __file__.replace('.py', '.png')
    except NameError:
        pass
    else:
        plot.save(outfile)
        print("Example output saved as\n%s" % outfile)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""`Spectrum` calculation methods using the SciPy module.
"""

from astropy import units

from .core import Spectrum
from .registry import register_method
from ..utils import import_method_dependency
from .utils import scale_timeseries_units
from .. import version
__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version


def welch(timeseries, segmentlength, noverlap=None, **kwargs):
    """Calculate the PSD using the scipy Welch method.
    """
    # get module
    signal = import_method_dependency('scipy.signal')
    # calculate PSD
    f, psd_ = signal.welch(timeseries.data, noverlap=noverlap,
                           fs=timeseries.sample_rate.decompose().value,
                           nperseg=segmentlength, **kwargs)
    # generate Spectrum and return
    spec = psd_.view(Spectrum)
    spec.frequencies = f
    spec.name = timeseries.name
    spec.epoch = timeseries.epoch
    spec.channel = timeseries.channel
    spec.unit = scale_timeseries_units(timeseries.unit,
                                       kwargs.get('scaling', 'density'))
    return spec

register_method(welch)


def bartlett(timeseries, segmentlength, **kwargs):
    """Calculate a PSD using the Bartlett average method.
    """
    import_method_dependency('scipy.signal')
    return welch(timeseries, segmentlength, noverlap=0, **kwargs)

register_method(bartlett)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""The `Series` is a one-dimensional array with metadata
"""

import numpy
import inspect
from scipy import signal

from astropy.units import (Unit, Quantity)

from .. import version
__version__ = version.version
__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

from .array import Array
from ..segments import Segment


class Series(Array):
    """A one-dimensional array with metadata
    """
    _metadata_slots = Array._metadata_slots + ['x0', 'dx', 'xunit', 'logx']
    xunit = Unit('')
    def __new__(cls, data, dtype=None, copy=False, subok=True, **metadata):
        """Define a new `Series`
        """
        if isinstance(data, (list, tuple)):
           data = numpy.asarray(data)
        if not data.ndim == 1:
            raise ValueError("Cannot create a %s with more than one "
                             "dimension" % cls.__name__)
        return super(Series, cls).__new__(cls, data, dtype=dtype, copy=copy,
                                          subok=subok, **metadata)

    # simple slice, need to copy x0 away from self
    def __getslice__(self, i, j):
        new = super(Series, self).__getslice__(i, j)
        new = new.copy()
        new.x0 = float(self.x0.value)
        new.x0 += (i * new.dx)
        return new

    # rebuild getitem to handle complex slicing
    def __getitem__(self, item):
        if isinstance(item, (float, int)):
            return Quantity(super(Series, self).__getitem__(item), self.unit)
        elif isinstance(item, slice):
            item = slice(item.start is not None and int(item.start) or None,
                         item.stop is not None and int(item.stop) or None,
                         item.step is not None and int(item.step) or None)
            new = super(Series, self).__getitem__(item)
            if item.start:
                new.x0 = float(self.x0.value)
                new.x0 += item.start * new.dx
            if item.step:
                new.dx = float(self.dx.value)
                new.dx *= item.step
            return new
        else:
            new = super(Series, self).__getitem__(item)
            if isinstance(item, (list, tuple, numpy.ndarray)):
                new.index = self.index[item]
            else:
                new.index = self.index[item.argmax()]
            return new

    # -------------------------------------------
    # Series properties

    @property
    def x0(self):
        """X-axis value of first sample
        """
        return self.metadata['x0']

    @x0.setter
    def x0(self, value):
        if isinstance(value, Quantity):
            self.metadata['x0'] = value.to(self.xunit)
        else:
            self.metadata['x0'] = Quantity(value, self.xunit)

    @x0.deleter
    def x0(self):
        del self.metadata['x0']

    @property
    def dx(self):
        """Distance between samples on the x-axis
        """
        return self.metadata['dx']

    @dx.setter
    def dx(self, value):
        if isinstance(value, Quantity):
            self.metadata['dx'] = value.to(self.xunit)
        else:
            self.metadata['dx'] = Quantity(value, self.xunit)

    @dx.deleter
    def dx(self):
        del self.metadata['dx']

    @property
    def span(self):
        """Extent of this `Series`
        """
        return Segment(self.x0.value,
                       self.x0.value + self.shape[0] * self.dx.value)

    @property
    def index(self):
        """Positions of the data on the x-axis

        :type: `Series`
        """
        try:
            return self._index
        except AttributeError:
            if not self.size:
                self.index = numpy.ndarray(0)
            elif self.logx:
                logdx = (numpy.log10(self.x0.value + self.dx.value) -
                         numpy.log10(self.x0.value))
                logx1 = numpy.log10(self.x0.value) + self.shape[-1] * logdx
                self.index = numpy.logspace(numpy.log10(self.x0.value), logx1,
                                             num=self.shape[-1])
            else:
                self.index = (numpy.arange(self.shape[-1]) * self.dx.value +
                              self.x0.value)
            return self.index

    @index.setter
    def index(self, samples):
        if not isinstance(samples, Array):
            fname = inspect.stack()[0][3]
            name = '%s %s' % (self.name, fname)
            samples = Array(samples, unit=self.xunit, name=name,
                            epoch=self.epoch, channel=self.channel, copy=True)
        self._index = samples
        try:
            self.x0 = self.index[0]
        except IndexError:
            pass
        try:
            self.dx = self.index[1] - self.index[0]
        except IndexError:
            pass

    @property
    def logx(self):
        """Boolean telling whether this `Series` has a logarithmic
        x-axis scale
        """
        try:
            return self.metadata['logx']
        except KeyError:
            self.logx = False
            return self.logx

    @logx.setter
    def logx(self, val):
        if (val and 'logx' in self.metadata and not self.metadata['logx'] and
                'index' in self.metadata):
            del self.index
        self.metadata['logx'] = bool(val)

    # -------------------------------------------
    # Series methods

    def resample(self, rate, window=None, dtype=None):
        """Resample this Series to a new rate

        Parameters
        ----------
        rate : `float`
            rate to which to resample this `Series`
        window : array_like, callable, string, float, or tuple, optional
            specifies the window applied to the signal in the Fourier
            domain.
        dtype : :class:`numpy.dtype`, default: `None`
            specific data type for output, defaults to input dtype

        Returns
        -------
        Series
            a new Series with the resampling applied, and the same
            metadata
        """
        if isinstance(rate, Quantity):
            rate = rate.value
        N = int(self.shape[0] * self.dx.value * rate)
        data = signal.resample(self.data, N, window=window)
        new = self.__class__(data, dtype=dtype or self.dtype)
        new.metadata = self.metadata.copy()
        new.dx = 1 / float(rate)
        return new

    def decimate(self, q, n=None, ftype='iir', axis=-1):
        """Downsample the signal by using a filter.

        By default, an order 8 Chebyshev type I filter is used.
        A 30 point FIR filter with hamming window is used if
        `ftype` is 'fir'.

        Parameters
        ----------
        q : `int`
            the downsampling factor.
        n : `int`, optional
            the order of the filter (1 less than the length for 'fir').
        ftype : `str`: {'iir', 'fir'}, optional
            the type of the lowpass filter.
        axis : `int`, optional
            The axis along which to decimate.

        Returns
        -------
        y : ndarray
        The down-sampled signal.

        See also
        --------
        :meth:`scipy.resample`
        """
        if not isinstance(q, int):
            raise TypeError("q must be an integer")
        if n is None:
            if ftype == 'fir':
                n = 30
            else:
                n = 8
        if ftype == 'fir':
            b = signal.firwin(n + 1, 1. / q, window='hamming')
            a = 1.
        else:
            b, a = signal.cheby1(n, 0.05, 0.8 / q)
        y = signal.lfilter(b, a, self.data, axis=axis)
        out = self.__class__(y, **self.metadata)
        sl = [slice(None)] * y.ndim
        sl[axis] = slice(None, None, q)
        return out[sl]

    # -------------------------------------------
    # numpy.ndarray method modifiers
    # all of these try to return Quantities rather than simple numbers

    def max(self, *args, **kwargs):
        return Quantity(super(Series, self).max(*args, **kwargs),
                        unit=self.unit)
    max.__doc__ = Array.max.__doc__

    def min(self, *args, **kwargs):
        return Quantity(super(Series, self).min(*args, **kwargs),
                        unit=self.unit)
    min.__doc__ = Array.min.__doc__

    def mean(self, *args, **kwargs):
        return Quantity(super(Series, self).mean(*args, **kwargs),
                        unit=self.unit)
    mean.__doc__ = Array.mean.__doc__

    def median(self, *args, **kwargs):
        return Quantity(super(Series, self).median(*args, **kwargs),
                        unit=self.unit)
    median.__doc__ = Array.median.__doc__



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit test for detector module
"""

import unittest

import numpy

from astropy import units

from .. import version
from ..detector import Channel
from ..utils import with_import

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version


class ChannelTests(unittest.TestCase):
    """`TestCase` for the timeseries module
    """
    channel = 'L1:PSL-ISS_PDB_OUT_DQ'

    def test_empty_create(self):
        new = Channel('')
        self.assertTrue(str(new) == '')
        self.assertTrue(new.sample_rate is None)
        self.assertTrue(new.dtype == numpy.float64)

    def test_create(self):
        new = Channel('L1:LSC-DARM_ERR', sample_rate=16384, unit='m')
        self.assertTrue(str(new) == 'L1:LSC-DARM_ERR')
        self.assertTrue(new.ifo == 'L1')
        self.assertTrue(new.system == 'LSC')
        self.assertTrue(new.subsystem == 'DARM')
        self.assertTrue(new.signal == 'ERR')
        self.assertTrue(new.sample_rate == units.Quantity(16384, 'Hz'))
        self.assertTrue(new.unit == units.meter)
        self.assertTrue(new.texname == r'L1:LSC-DARM\_ERR')

    def test_query(self):
        try:
            new = Channel.query(self.channel)
        except Exception as e:
            try:
                import kerberos
            except ImportError:
                raise unittest.SkipTest('Channel.query() requires kerberos '
                                         'to be installed')
            else:
                if isinstance(e, kerberos.GSSError):
                    raise unittest.SkipTest(str(e))
                else:
                    raise
        self.assertTrue(str(new) == self.channel)
        self.assertTrue(new.ifo == self.channel.split(':', 1)[0])
        self.assertTrue(new.sample_rate == units.Quantity(32768, 'Hz'))

        #self.assertTrue(new.unit == units.meter)
        self.assertTrue(new.texname == self.channel.replace('_', r'\_'))

    def test_nds2_conversion(self):
        try:
            import nds2
        except ImportError as e:
            return unittest.SkipTest(str(e))
        else:
            try:
                conn = nds2.connection('nds2.ligo-la.caltech.edu')
            except Exception as f:
                return unittest.SkipTest(str(f))
            else:
                nds2channel = conn.find_channels(self.channel)[0]
                new = Channel.from_nds2(nds2channel)
                self.assertTrue(str(new) == self.channel)
                self.assertTrue(new.ifo == self.channel.split(':', 1)[0])
                self.assertTrue(new.sample_rate == units.Quantity(32768, 'Hz'))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Decorator for GWpy plotting
"""

import threading

from matplotlib.figure import Figure
from matplotlib.axes import Axes

from .decorator import decorator

from .. import version
__version__ = version.version
__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

mydata = threading.local()


@decorator
def auto_refresh(f, *args, **kwargs):
    refresh = kwargs.pop('refresh', True)
    # The following is necessary rather than using mydata.nesting = 0 at the
    # start of the file, because doing the latter caused issues with the Django
    # development server.
    mydata.nesting = getattr(mydata, 'nesting', 0) + 1
    try:
        return f(*args, **kwargs)
    finally:
        mydata.nesting -= 1
        if isinstance(args[0], Axes):
            if refresh and mydata.nesting == 0 and args[0].figure._auto_refresh:
                args[0].figure.refresh()
        elif isinstance(args[0], Figure):
            if refresh and mydata.nesting == 0 and args[0]._auto_refresh:
                args[0].refresh()

@decorator
def axes_method(f, *args, **kwargs):
    figure = args[0]
    axes = [ax for ax in figure.axes if ax not in figure._coloraxes]
    if len(axes) == 0:
        raise RuntimeError("No axes found for which '%s' is applicable"
                           % f.__name__)
    if len(axes) != 1:
        raise RuntimeError("{0} only applicable for a Plot with a single set "
                           "of data axes. With multiple data axes, you should "
                           "access the {0} method of the relevant Axes (stored "
                           "in ``Plot.axes``) directly".format(f.__name__))
    axesf = getattr(axes[0], f.__name__)
    return axesf(*args[1:], **kwargs)
